# -*- coding: utf-8 -*-

"""Custom `inspect` module."""

__author__ = "Nathaniel Starkman"


##############################################################################
# IMPORTS

# BUILT-IN
import inspect
import typing as T
from collections import namedtuple
from inspect import *  # noqa  # so can be a drop-in for `inspect`
from inspect import FullArgSpec, Parameter
from inspect import Signature as Signature
from inspect import _void, getfullargspec

# THIRD PARTY
from astropy.utils.decorators import format_doc
from typing_extensions import Literal

##############################################################################
# PARAMETERS

__all__ = [
    "POSITIONAL_ONLY",
    "POSITIONAL_OR_KEYWORD",
    "VAR_POSITIONAL",
    "KEYWORD_ONLY",
    "VAR_KEYWORD",
    # "_void",  # TODO: add to RTD
    # "_empty",  # TODO: add to RTD
    "_placehold",
    "_is_empty",
    "_is_void",
    "_is_placehold",
    "_is_placeholder",
    "FullerArgSpec",
    "getfullerargspec",
    "get_annotations_from_signature",
    "get_defaults_from_signature",
    "get_kwdefaults_from_signature",
    "get_kwonlydefaults_from_signature",
    "get_kinds_from_signature",
    "modify_parameter",
    "replace_with_parameter",
    "insert_parameter",
    "prepend_parameter",
    "append_parameter",
    "drop_parameter",
    "FullerSignature",
    "fuller_signature",
]

POSITIONAL_ONLY = Parameter.POSITIONAL_ONLY
POSITIONAL_OR_KEYWORD = Parameter.POSITIONAL_OR_KEYWORD
VAR_POSITIONAL = Parameter.VAR_POSITIONAL
KEYWORD_ONLY = Parameter.KEYWORD_ONLY
VAR_KEYWORD = Parameter.VAR_KEYWORD

# placeholders
_empty = Parameter.empty  # TODO: add to RTD
_void = _void  # TODO: add to RTD


# class _void:  # FIX import _void from `inspect`
#     """Void."""

#     pass


# /class


class _placehold:
    """Placehold."""


# /class


# types

# EmptyType = T.Type[_empty]
_typing_tuple_false = T.Union[tuple, Literal[False]]


FullerArgSpec: namedtuple = namedtuple(
    "FullerArgSpec",
    [
        "args",
        "defaultargs",
        "argdefaults",
        "varargs",
        "kwonlyargs",
        "kwonlydefaults",
        "varkw",
        "annotations",
        "docstring",
    ],
)


##############################################################################
# CODE
##############################################################################


##########################################################################
# safe placeholder comparison
# some 3rd party packages have quantities which cannot be directly compared to
# via ``==`` or ``!=`` and will throw an exception. These methods implement
# safe testing against ``_empty``, ``_void``, ``_placehold``
# and the combination.


def _is_empty(value):
    """Test whether `value`==`_empty`."""
    try:
        value == _empty
    except Exception:
        # if it throws an exception, it clearly isn't `_empty`
        return False
    else:
        return value == _empty


# /def


def _is_void(value):
    """Test whether `value`==`_void`."""
    try:
        value == _void
    except Exception:
        # if it throws an exception, it clearly isn't `_void`
        return False
    else:
        return value == _void


# /def


def _is_placehold(value):
    """Test whether `value`==`_placehold`."""
    try:
        value == _placehold
    except Exception:
        # if it throws an exception, it clearly isn't `_placehold`
        return False
    else:
        return value == _placehold


# /def


def _is_placeholder(value):
    """Test whether `value`==`_placeholder`."""
    try:
        value == _empty
    except Exception:
        # if it throws an exception, it clearly isn't `_placeholder`
        return False
    else:
        return (value == _empty) | (value == _void) | (value == _placehold)


# /def


###########################################################################
# getfullerargspec


def getfullerargspec(func: T.Callable) -> FullerArgSpec:
    """Separated version of FullerArgSpec.

    fullargspec with separation of mandatory and optional arguments
    adds *defargs* which corresponds *defaults*

    Parameters
    ----------
    func : function
        the function to inspect

    Returns
    -------
    FullerArgSpec : namedtuple
        args             : the mandatory arguments
        defargs          : arguments with defaults
        defaults         : dictionary of defaults to `defargs`
        varargs          : variable arguments (args)
        kwonlyargs       : keyword-only arguments
        kwonlydefaults   : keyword-only argument defaults
        varkw            : variable key-word arguments (kwargs)
        annotations      : function annotations
        docstring        : function docstring

    """
    spec: FullArgSpec = getfullargspec(func)  # get argspec

    if spec.defaults is not None:  # separate out argument types

        args: T.Optional[T.List[str]] = spec.args[: -len(spec.defaults)]
        defargs: T.Optional[T.List[str, T.Any]] = spec.args[
            -len(spec.defaults) :
        ]
        defaults = {k: v for k, v in zip(defargs, spec.defaults)}

    else:  # nothing to separate
        args = spec.args
        defargs = None
        defaults = None

    # build FullerArgSpec
    return FullerArgSpec(
        args=args,
        defaultargs=defargs,
        argdefaults=defaults,
        varargs=spec.varargs,
        kwonlyargs=spec.kwonlyargs,
        kwonlydefaults=spec.kwonlydefaults,
        varkw=spec.varkw,
        annotations=spec.annotations,
        docstring=func.__doc__,
    )


# /def


###########################################################################
# Signature / ArgSpec Interface


def get_annotations_from_signature(signature: Signature) -> T.Dict[str, T.Any]:
    """Get annotations from Signature object.

    Parameters
    ----------
    signature: Signature
        the object's signature

    Returns
    -------
    annotations: dict
        argument {name: annotation} values
        return annotations under key 'return'

    Examples
    --------
    >>> def func(x: 'x annotation') -> 'return annotation':
    ...   pass
    >>> sig = Signature.from_callable(func)
    >>> get_annotations_from_signature(sig)
    {'x': 'x annotation', 'return': 'return annotation'}

    """
    annotations: T.Dict[str, T.Any] = {
        k: v.annotation
        for k, v in signature.parameters.items()
        if not _is_empty(v.annotation)
    }
    if not _is_empty(signature.return_annotation):
        annotations["return"] = signature.return_annotation

    return annotations


# /def


def get_defaults_from_signature(signature: Signature) -> tuple:
    """Get defaults from Signature object.

    Parameters
    ----------
    signature: Signature
        the object's signature

    Returns
    -------
    defaults: tuple
        n-tuple for n defaulted positional parameters

    Examples
    --------
    >>> def func(x=2,):
    ...     pass
    >>> FullerSignature.from_callable(func).defaults
    (2,)

    this does not get the keyword only defaults

    >>> def func(x=2,*,k=3):
    ...     pass
    >>> FullerSignature.from_callable(func).defaults
    (2,)

    """
    return tuple(
        [
            p.default
            for p in signature.parameters.values()
            if (
                (p.kind == POSITIONAL_OR_KEYWORD) & ~_is_empty(p.default)
            )  # the kind
        ]
    )  # only defaulted


# /def


def get_kwdefaults_from_signature(signature: Signature) -> dict:
    """Get keyword-only defaults from Signature object.

    Parameters
    ----------
    signature: Signature
        the object's signature

    Returns
    -------
    defaults: dict
        argument {name: default}

    Examples
    --------
    >>> def func(x=2,*,k=3):
    ...     pass
    >>> FullerSignature.from_callable(func).kwdefaults
    {'k': 3}

    """
    return {
        n: p.default
        for n, p in signature.parameters.items()
        if ((p.kind == KEYWORD_ONLY) and not _is_empty(p.default))
    }


# /def


get_kwonlydefaults_from_signature = get_kwdefaults_from_signature


def get_kinds_from_signature(signature: Signature) -> tuple:
    """Get parameter kinds from Signature object.

    Parameters
    ----------
    signature: Signature
        the object's signature

    Returns
    -------
    kinds: tuple
        POSITIONAL_OR_KEYWORD, VAR_POSITIONAL, KEYWORD_ONLY, VAR_KEYWORD

    Examples
    --------
    >>> def func(x, *args, k=3, **kw):
    ...     pass
    >>> kinds = FullerSignature.from_callable(func).kinds
    >>> kinds[0]
    <_ParameterKind.POSITIONAL_OR_KEYWORD: 1>
    >>> kinds[1]
    <_ParameterKind.VAR_POSITIONAL: 2>
    >>> kinds[2]
    <_ParameterKind.KEYWORD_ONLY: 3>
    >>> kinds[3]
    <_ParameterKind.VAR_KEYWORD: 4>

    """
    return tuple([p.kind for p in signature.parameters.values()])


# /def


###########################################################################
# Signature Methods


def modify_parameter(
    sig: Signature,
    param: T.Union[str, int],
    name: T.Union[str, _empty] = _empty,
    kind: T.Any = _empty,
    default: T.Any = _empty,
    annotation: T.Any = _empty,
) -> Signature:
    """Modify a Parameter.

    Similar to `.replace,` but more convenient for modifying a single parameter
    Parameters are immutable, so will create a new `Signature` object

    Parameters
    ----------
    sig:  Signature
        Signature object
    param: int or str
        the parameter index (or name) in `self.parameters`
    name: str
        new parameter name, defaults to old parameter name
        **default: None**
    kind: type
        new parameter kind, defaults to old parameter kind
        **default: None**
    default: any
        new parameter default, defaults to old parameter default
        **default: None**
    annotation: any
        new parameter annotation, defaults to old parameter annotation
        **default: None**

    Returns
    -------
    Signature
        a new Signature object with the replaced parameter

    """
    # identify parameter to modify
    if isinstance(param, int):
        index = param
    else:
        index = list(sig.parameters.keys()).index(param)
    # get the parameters, and the specific param
    params = list(sig.parameters.values())
    _param = params[index]

    # replacements
    name = _param.name if name is _empty else name
    kind = _param.kind if kind is _empty else kind
    default = _param.default if default is _empty else default
    annotation = _param.annotation if annotation is _empty else annotation

    # adjust parameter list
    params[index] = _param.replace(
        name=name, kind=kind, default=default, annotation=annotation
    )

    return sig.replace(parameters=params)


# /def


def replace_with_parameter(
    sig: Signature, name: T.Union[int, str], param: Parameter
) -> Signature:
    """Replace a Parameter with another Parameter.

    Similar to `.replace,` but more convenient for modifying a single parameter
    Parameters are immutable, so will create a new `Signature` object

    Parameters
    ----------
    sig:  Signature
        Signature object
    name: int or str
        parameter to replace
    param: Parameter
        new parameter kind, defaults to old parameter kind
        **default: None**

    Returns
    -------
    Signature
        a new Signature object with the replaced parameter

    """
    # identify parameter to replace
    if isinstance(name, int):  # convert index to name
        index = name
        name = list(sig.parameters.keys())[name]
    else:
        index = list(sig.parameters.keys()).index(name)

    sig = drop_parameter(sig, name)
    sig = insert_parameter(sig, index, param)

    return sig


# /def


def insert_parameter(
    sig: Signature, index: int, param: Parameter
) -> Signature:
    """Insert a new Parameter.

    Similar to .replace, but more convenient for adding a single parameter
    Parameters are immutable, so will create a new Signature object

    Parameters
    ----------
    sig:  Signature
        Signature object
    index: int
        index into Signature.parameters at which to insert new parameter
    param: Parameter
        param to insert at index

    Returns
    -------
    Signature:
        a new Signature object with the inserted parameter

    """
    parameters = list(sig.parameters.values())
    parameters.insert(index, param)

    return sig.replace(parameters=parameters)


# /def


def prepend_parameter(sig: Signature, param: Parameter) -> Signature:
    """Insert a new Parameter at the start.

    Similar to .replace, but more convenient for adding a single parameter
    Parameters are immutable, so will create a new Signature object

    Parameters
    ----------
    sig:  Signature
        Signature object
    index: int
        index into Signature.parameters at which to insert new parameter
    param: Parameter
        param to insert at `index`

    Returns
    -------
    Signature: Signature
        a new `Signature` object with the inserted `param`

    .. todo::

        have a `skip_self` option to skip self/cls in class methods.

    """
    return insert_parameter(sig, 0, param)


# /def


def append_parameter(sig: Signature, param: Parameter) -> Signature:
    """Insert a new Parameter at the end.

    Similar to .replace, but more convenient for adding a single parameter
    Parameters are immutable, so will create a new Signature object

    Parameters
    ----------
    sig:  Signature
        Signature object
    index: int
        index into Signature.parameters at which to insert new parameter
    param: Parameter
        param to insert at `index`

    Returns
    -------
    Signature: Signature
        a new `Signature` object with the inserted `param`

    """
    kinds = get_kinds_from_signature(sig)
    return insert_parameter(sig, len(kinds) + 1, param)


# /def


def drop_parameter(
    sig: Signature, param: T.Union[str, int, Parameter, None]
) -> Signature:
    """Drop a Parameter.

    Parameters
    ----------
    sig : Signature
        Signature object
    param: str, int, Parameter
        the parameter to drop in self.parameters
        identified by either the name (str) or index (int)
        (Parameter type calls name)
        If None, does not drop anything

    Returns
    -------
    Signature:
        a new Signature object with the replaced parameter

    """
    if param is None:
        return sig
    elif isinstance(param, int):  # convert index to name
        index = param
    elif isinstance(param, str):
        index = list(sig.parameters.keys()).index(param)
    elif isinstance(param, Parameter):
        index = list(sig.parameters.keys()).index(param.name)
    else:
        raise TypeError
    # setup
    parameters = list(sig.parameters.values())

    # drop
    del parameters[index]

    return sig.replace(parameters=parameters)


# /def


###########################################################################
# Signature


class FullerSignature(Signature):
    """Signature with better ArgSpec compatibility.

    Constructs FullerSignature from the given list of Parameter
    objects and 'return_annotation'.  All arguments are optional.

    Though `Signature` is the new object, python still largely
    uses the outputs  as  defined by ``getfullargspec``
    This serves as a bridge, providing methods that return
    the same output as ``getfullargspec``

    Parameters
    ----------
    parameters : Sequence, optional
        list of Parameter objects
    return_annotation : Any
        return annotation of `obj`
    obj : Any
        the object for which this is the signature

    """

    def __init__(
        self,
        parameters=None,
        *,
        return_annotation=_empty,
        obj=None,
        __validate_parameters__=True
    ):
        super().__init__(
            parameters=parameters,
            return_annotation=return_annotation,
            __validate_parameters__=__validate_parameters__,
        )

        self.obj = obj

    # /def

    @classmethod
    def from_callable(cls, obj, *, follow_wrapped=True):
        """From callable.

        Parameters
        ----------
        obj : Callable
        follow_wrapped : bool

        Returns
        -------
        FullerSignature

        """
        sig = super().from_callable(obj, follow_wrapped=follow_wrapped)
        sig = FullerSignature(
            parameters=sig.parameters.values(),
            return_annotation=sig.return_annotation,
            obj=obj,
        )
        return sig

    # /def

    @classmethod
    def from_signature(cls, signature: Signature, *, obj=None):
        """Create :class:`FullerSignature` from :class:`~inspect.Signature`.

        Parameters
        ----------
        signature : Signature

        Returns
        -------
        FullerSignature instance

        """
        sig = cls(
            parameters=signature.parameters.values(),
            return_annotation=signature.return_annotation,
            obj=obj,
        )

        # TODO check on obj, that has matching sig as signature

        return sig

    # /def

    @format_doc(None, original_doc=Signature.bind.__doc__)
    def bind_with_defaults(self, *args, **kwargs):
        """Bind arguments to parameters.

        {original_doc}

        Applies defaults to the BoundArgument using ``apply_defaults``.

        """
        ba = self.bind(*args, **kwargs)
        ba.apply_defaults()
        return ba

    # /def

    @format_doc(None, original_doc=Signature.bind_partial.__doc__)
    def bind_partial_with_defaults(self, *args, **kwargs):
        """Bind (allowing omissions) arguments to parameters.

        {original_doc}

        Applies defaults to the BoundArgument using ``apply_defaults``.

        """
        ba = self.bind_partial(*args, **kwargs)
        ba.apply_defaults()
        return ba

    # /def

    # ------------------------------------------

    @property
    def __signature__(self) -> Signature:
        """Return a classical Signature."""
        return Signature(
            parameters=list(self.parameters.values()),
            return_annotation=self.return_annotation,
            # __validate_parameters__=False,
        )

    # /def

    @property
    def signature(self) -> Signature:
        """Return a classical Signature."""
        return self.__signature__

    # /def

    @property
    def __annotations__(self) -> dict:
        """Get annotations from Signature object.

        Returns
        -------
        annotations: dict
            argument {name: annotation} values
            return annotations under key 'return'

        Examples
        --------
        >>> def func(x: 'x annotation') -> 'return annotation':
        ...   pass
        >>> FullerSignature.from_callable(func).annotations
        {'x': 'x annotation', 'return': 'return annotation'}

        """
        annotations: T.Dict[str, T.Any] = get_annotations_from_signature(
            self.signature
        )

        return annotations

    # /def

    @property
    def annotations(self) -> T.Any:
        """Get annotations from Signature object.

        Returns
        -------
        annotations: dict
            argument {name: annotation} values
            return annotations under key 'return'

        Examples
        --------
        >>> def func(x: 'x annotation') -> 'return annotation':
        ...   pass
        >>> FullerSignature.from_callable(func).annotations
        {'x': 'x annotation', 'return': 'return annotation'}

        """
        return self.__annotations__

    # /def

    @property
    def __defaults__(self) -> T.Optional[tuple]:
        """Get defaults.

        Returns
        -------
        tuple
            n-tuple for n defaulted positional parameters

        Examples
        --------
        >>> def func(x=2,):
        ...     pass
        >>> FullerSignature.from_callable(func).defaults
        (2,)

        """
        return get_defaults_from_signature(self.signature)

    # /def

    @property
    def defaults(self) -> T.Optional[tuple]:
        """Get defaults.

        Returns
        -------
        tuple
            n-tuple for n defaulted positional parameters

        Examples
        --------
        >>> def func(x=2,):
        ...     pass
        >>> FullerSignature.from_callable(func).defaults
        (2,)

        """
        return self.__defaults__

    # /def

    @property
    def __kwdefaults__(self) -> T.Optional[dict]:
        """Get keyword-only defaults.

        Returns
        -------
        defaults: dict
            argument {name: default value}

        Examples
        --------
        >>> def func(x=2,*,k=3):
        ...     pass
        >>> FullerSignature.from_callable(func).kwdefaults
        {'k': 3}

        """
        return get_kwdefaults_from_signature(self.signature)

    # /def

    @property
    def kwdefaults(self) -> T.Optional[dict]:
        """Get keyword-only defaults.

        Returns
        -------
        defaults: dict
            argument {name: default}

        Examples
        --------
        >>> def func(x=2,*,k=3):
        ...     pass
        >>> FullerSignature.from_callable(func).kwdefaults
        {'k': 3}

        """
        return self.__kwdefaults__

    # /def

    @property
    def kwonlydefaults(self) -> T.Optional[dict]:
        """Get keyword-only defaults.

        Returns
        -------
        defaults: dict
            argument {name: default}

        Examples
        --------
        >>> def func(x=2,*,k=3):
        ...     pass
        >>> FullerSignature.from_callable(func).kwdefaults
        {'k': 3}

        """
        return self.__kwdefaults__

    # /def

    @property
    def kinds(self) -> tuple:
        """Get parameter kinds.

        Returns
        -------
        kinds: tuple
            POSITIONAL_OR_KEYWORD, VAR_POSITIONAL, KEYWORD_ONLY, VAR_KEYWORD

        Examples
        --------
        >>> def func(x, *args, k=3, **kw):
        ...     pass
        >>> kinds = FullerSignature.from_callable(func).kinds
        >>> kinds[0]
        <_ParameterKind.POSITIONAL_OR_KEYWORD: 1>
        >>> kinds[1]
        <_ParameterKind.VAR_POSITIONAL: 2>
        >>> kinds[2]
        <_ParameterKind.KEYWORD_ONLY: 3>
        >>> kinds[3]
        <_ParameterKind.VAR_KEYWORD: 4>

        """
        return get_kinds_from_signature(self.signature)

    @property
    def names(self) -> tuple:
        """Get parameter kinds.

        Returns
        -------
        names: tuple of str

        Examples
        --------
        >>> def func(x, *args, k=3, **kw):
        ...     pass
        >>> FullerSignature.from_callable(func).names
        ('x', 'args', 'k', 'kw')

        """
        return tuple(self.parameters.keys())

    # /def

    #     (args, varargs, varkw, defaults, kwonlyargs, kwonlydefaults, annotations).
    #     'args' is a list of the parameter names.
    #     'varargs' and 'varkw' are the names of the * and ** parameters or None.
    #     'defaults' is an n-tuple of the default values of the last n parameters.
    #     'kwonlyargs' is a list of keyword-only parameter names.
    #     'kwonlydefaults' is a dictionary mapping names from kwonlyargs to defaults.
    #     'annotations' is a dictionary mapping parameter names to annotations.

    #     def fullargspec(self):
    #         return

    # ------------------------------------------

    @property
    def index_positional(self) -> _typing_tuple_false:
        """Index(ices) of positional arguments.

        This includes defaulted positional arguments.

        Returns
        -------
        tuple or False
            False if no positional arguments, tuple of indices otherwise

        """
        kinds: tuple = self.kinds
        try:
            kinds.index(1)  # POSITIONAL_OR_KEYWORD = 1
        except ValueError:
            return False
        else:
            return tuple(
                [i for i, k in enumerate(kinds) if ((k == 0) | (k == 1))]
            )

    # /def

    @property
    def index_positional_only(self) -> _typing_tuple_false:
        """Index(ices) of positional-only arguments.

        Returns
        -------
        tuple or False
            False if no positional arguments, tuple of indices otherwise

        """
        kinds: tuple = self.kinds
        try:
            kinds.index(0)  # POSITIONAL_ONLY = 0
        except ValueError:
            return False
        else:
            return tuple([i for i, k in enumerate(kinds) if (k == 0)])

    # /def

    @property
    def index_positional_defaulted(self) -> _typing_tuple_false:
        """Index(ices) of positional arguments with default values.

        FIXME, wrong b/c returns indices of POSITIONAL_OR_KEYWORD arguments
        that do not have defaults

        Returns
        -------
        tuple or False
            False if no positional arguments, tuple of indices otherwise

        """
        kinds: tuple = self.kinds
        try:
            kinds.index(1)  # POSITIONAL_OR_KEYWORD = 1
        except ValueError:
            return False
        else:
            pos_only: list = self.index_positional_only or []

            return tuple(
                [
                    i
                    for i, k in enumerate(kinds)
                    if ((k == 1) & (i not in pos_only))
                ]
            )

    # /def

    @property
    def index_var_positional(self) -> T.Union[int, Literal[False]]:
        """Index of `*args`.

        Returns
        -------
        int or False
            False if no variable positional argument, index int otherwise

        """
        kinds = self.kinds
        try:
            kinds.index(2)  # VAR_POSITIONAL = 2
        except ValueError:
            return False
        else:
            return kinds.index(2)

    # /def

    @property
    def index_keyword_only(self) -> _typing_tuple_false:
        """Index of `*args`.

        Returns
        -------
        tuple or False
            False if no keyword-only arguments, tuple of indices otherwise

        """
        kinds = self.kinds
        try:
            kinds.index(3)  # KEYWORD_ONLY = 3
        except ValueError:
            return False
        else:
            return tuple([i for i, k in enumerate(kinds) if (k == 3)])

    # /def

    @property
    def index_end_keyword_only(self) -> int:
        """Index to place new keyword-only parameters.

        Returns
        -------
        int
            var_keyword index if var_keyword exists, last index otherwise

        """
        index = self.index_var_keyword
        if index is False:  # no variable kwargs
            index = len(self.kinds) + 1
        return index

    # /def

    @property
    def index_var_keyword(self) -> T.Union[int, Literal[False]]:
        """Index of `**kwargs`.

        Returns
        -------
        int or False
            False if no variable keyword argument, index int otherwise

        """
        kinds = self.kinds
        try:
            kinds.index(4)  # VAR_KEYWORD = 4
        except ValueError:
            return False
        else:
            return kinds.index(4)

    # /def

    # ------------------------------------------

    def copy(self):
        """Copy of self."""
        return self.replace(parameters=list(self.parameters.values()))

    # ------------------------------------------

    def modify_parameter(
        self,
        param: T.Union[str, int],
        name: T.Union[str, _empty] = _empty,
        kind: T.Any = _empty,
        default: T.Any = _empty,
        annotation: T.Any = _empty,
    ):
        """Modify a Parameter.

        Similar to `.replace,` but more convenient for modifying a single parameter
        Parameters are immutable, so will create a new `Signature` object

        Parameters
        ----------
        param: int or str
            the parameter index (or name) in `self.parameters`
        name: str
            new parameter name, defaults to old parameter name
            **default: None**
        kind: type
            new parameter kind, defaults to old parameter kind
            **default: None**
        default: T.Any
            new parameter default, defaults to old parameter default
            **default: None**
        annotation: T.Any
            new parameter annotation, defaults to old parameter annotation
            **default: None**

        Returns
        -------
        FullerSignature
            a new Signature object with the replaced parameter

        """
        return modify_parameter(
            self,
            param=param,
            name=name,
            kind=kind,
            default=default,
            annotation=annotation,
        )

    # /def

    def replace_with_parameter(
        self, name: T.Union[int, str], param: Parameter
    ):
        """Replace a Parameter with another Parameter.

        Similar to `.replace,` but more convenient for modifying a single parameter
        Parameters are immutable, so will create a new `Signature` object

        Parameters
        ----------
        name: int or str
            parameter to replace
        param: Parameter
            new parameter kind, defaults to old parameter kind
            **default: None**

        Returns
        -------
        FullerSignature
            a new Signature object with the replaced parameter

        """
        return replace_with_parameter(self, name, param)

    # /def

    def insert_parameter(self, index: int, parameter: Parameter):
        """Insert a new Parameter.

        Similar to .replace, but more convenient for adding a single parameter
        Parameters are immutable, so will create a new Signature object

        Parameters
        ----------
        index: int
            index into Signature.parameters at which to insert new parameter
        parameter: Parameter
            parameter to insert at `index`

        Returns
        -------
        FullerSignature
            a new `Signature` object with the inserted `parameter`

        """
        return insert_parameter(self, index, parameter)

    # /def

    def prepend_parameter(self, param: Parameter):
        """Insert a new Parameter at the start.

        Similar to .replace, but more convenient for adding a single parameter
        Parameters are immutable, so will create a new Signature object

        Parameters
        ----------
        index: int
            index into Signature.parameters at which to insert new parameter
        param: Parameter
            param to insert at `index`

        Returns
        -------
        FullerSignature
            a new `Signature` object with the inserted `param`

        .. todo::

            have a `skip_self` option to skip self/cls in class methods.

        """
        return prepend_parameter(self, param)

    # /def

    def append_parameter(self, param: Parameter):
        """Insert a new Parameter at the end.

        Similar to .replace, but more convenient for adding a single parameter
        Parameters are immutable, so will create a new Signature object

        Parameters
        ----------
        index: int
            index into Signature.parameters at which to insert new parameter
        param: Parameter
            param to insert at `index`

        Returns
        -------
        FullerSignature
            a new `Signature` object with the inserted `param`

        """
        return append_parameter(self, param)

    # /def

    def drop_parameter(self, param: str):
        """Drop a Parameter.

        Parameters
        ----------
        param: str
            the parameter name in self.parameters

        Returns
        -------
        FullerSignature
            a new FullerSignature object with the replaced parameter

        """
        return drop_parameter(self, param)

    # /def

    def _default_pos_to_kwonly_from(self, index: int = 0):
        """Promote default positional to keyword only arguments.

        Parameter
        ---------
        index: int, optional
            default positional arguments after `index`
            will be changed to keyword only arguments
            ``None`` changes all available arguments

            example, (x, y=2, z=3) with index 1 means only z is changed
            because the defaults are (2, 3)

        Returns
        -------
        Signature
            a new Signature object with the promoted parameters

        TODO
        ----
        return list/info of parameters promoted

        """
        signature: FullerSignature = self

        if signature.defaults is None:  # no defaults
            return signature

        # promote parameters
        i: int
        for i in signature.index_positional_defaulted[index:][::-1]:
            signature = signature.modify_parameter(i, kind=KEYWORD_ONLY)

        return signature

    # /def

    def add_var_positional_parameter(
        self, name: str = "args", index: T.Optional[int] = None
    ):
        """Add var positional parameter.

        Does not add if one already exists.

        Parameters
        ----------
        name: str
            the var_positional argument name
            (default 'args')
        index: int, optional
            the index at which to place the var_positional_parameter
            default positional arguments after the var_positional parameter
            will be changed to keyword only arguments

        Returns
        -------
        Signature
            a new Signature object with the new var_positional parameter
            and also promoted parameters if `promote_default_pos`=True

        """
        signature: FullerSignature = self

        if self.index_var_positional is not False:  # already exists
            pass

        else:  # doesn't have ``*``

            if index is not None:
                # promote default-valued positional arguments to kwargs
                if index in self.index_positional_defaulted:
                    pos_index = self.index_positional_defaulted.index(index)
                elif index < self.index_positional_defaulted[0]:
                    pos_index = None
                else:
                    pos_index = self.index_positional_defaulted[-1] + 1

                signature = self._default_pos_to_kwonly_from(index=pos_index)

            else:  # has no defaulted positionals
                if not self.index_positional:  # no positional
                    index = 0
                else:
                    index = self.index_positional[-1] + 1

            signature = signature.insert_parameter(
                index,
                Parameter(name, VAR_POSITIONAL),
            )

        return signature

    # /def

    def add_var_keyword_parameter(self, name: str = "kwargs"):
        """Add var keyword parameter.

        Does not add if one already exists.

        Parameters
        ----------
        name: str
            the var_keyword argument name
            (default 'kwargs')

        Returns
        -------
        FullerSignature
            a new Signature object with the new var_positional parameter
            and also promoted parameters if `promote_default_pos`=True

        """
        signature: FullerSignature

        if self.index_keyword_only is not False:  # already exists
            signature = self

        else:
            signature = self.append_parameter(
                Parameter(name, VAR_KEYWORD),
            )

        return signature

    # /def


# ------------------------------------------------------------------------


def fuller_signature(obj: T.Callable, *, follow_wrapped: bool = True):
    """Get a signature object for the passed callable."""
    return FullerSignature.from_callable(obj, follow_wrapped=follow_wrapped)


# /def


###########################################################################
# Signature Convenience Methods


def fuller_signature_from_method(method: T.Callable):
    sig = fuller_signature(method)

    if tuple(sig.parameters.keys())[0] == "self":
        sig = sig.drop_parameter("self")
    elif tuple(sig.parameters.keys())[0] == "cls":
        sig = sig.drop_parameter("cls")

    return sig


# /def


def signature_from_method(method: T.Callable):
    sig = inspect.signature(method)

    if tuple(sig.parameters.keys())[0] == "self":
        sig = drop_parameter(sig, "self")
    elif tuple(sig.parameters.keys())[0] == "cls":
        sig = sig.drop_parameter("cls")

    return sig


# /def


##############################################################################
# END



# -*- coding: utf-8 -*-

"""Warnings and Exceptions."""

__author__ = "Nathaniel Starkman"

__all__ = [
    "conf",
    # warning classes
    "utilipyWarning",
    "utilipyWarningVerbose",
    # functions
    "showwarning",
]

##############################################################################
# IMPORTS

# BUILT-IN
import sys
import typing as T
import warnings

# THIRD PARTY
from astropy import config as _config
from astropy.utils.exceptions import AstropyWarning

#############################################################################
# CONFIGURATION


class Conf(_config.ConfigNamespace):
    """Configuration parameters for :mod:`~utilipy.utils.exceptions`."""

    verbose_warnings = _config.ConfigItem(
        True,
        description="When True, use verbose `utilipy` warnings",
        cfgtype="boolean(default=True)",
    )


conf = Conf()


###############################################################################
# CODE
###############################################################################


class utilipyWarning(AstropyWarning):
    """:mod:`~utilipy` package warning."""


# /class


# ----------------------------------------------------------------------------


class utilipyWarningVerbose(utilipyWarning):
    """Verbose :mod:`~utilipy` warning.

    If used, the warning is verbose, regardless of the `conf` setting.

    """


# /class


# ----------------------------------------------------------------------------
# override built-in showwarning


def showwarning(
    message: T.Any,
    category: type = utilipyWarning,
    filename: str = "",
    lineno: int = -1,
    file: None = None,
    line: None = None,
):
    """Override :func:`~warnings.showwarning`.

    showwarning docs

        Write a warning to a file. The default implementation calls
        :func:`~warnings.formatwarning`(`message`, `category`, `filename`,
        `lineno`, `line`) and writes the resulting string to file, which
        defaults to :func:`~sys.stderr`. You may replace this function with
        any callable by assigning to :func:`~warnings.showwarning`. line is a
        line of source code to be included in the warning message; if line is
        not supplied, :func:`~warnings.formatwarning` will try to read the
        line specified by `filename` and `lineno`.

    Parameters
    ----------
    message : Any
    category : type, optional
    filename : str, optional
    lineno : int, optional
    file : file-type or None, optional
    line : str or None, optional

    """
    if file is None:
        file = sys.stderr

    # Three conditions for a non-verbose warning:
    #   1) utilipyWarning (or subclass)
    #   2) NOT utilipyWarningVerbose
    #   3) verbosity setting is OFF
    if (
        issubclass(category, utilipyWarning)
        and not issubclass(category, utilipyWarningVerbose)
        and not conf.verbose_warnings
    ):
        print(
            "utilipyWarning: " + str(message),
            file=file,
            flush=True,
        )

    # Normal warnings
    else:
        print(
            warnings.formatwarning(
                message=message,
                category=category,
                filename=filename,
                lineno=lineno,
                line=line,
            ),
            file=file,
            flush=True,
        )


# /def

warnings.showwarning = showwarning


##############################################################################
# END



# -*- coding: utf-8 -*-

"""Selection Functions.

Examples
--------
do some examples here

"""

__author__ = "Nathaniel Starkman"


__all__ = [
    "inRange",
    "outRange",
    "ioRange",
    "ellipse",
    "circle",
]


#############################################################################
# IMPORTS

# BUILT-IN
import typing as T

# THIRD PARTY
import numpy as np

# PROJECT-SPECIFIC
from .decorators import idxDecorator
from utilipy.utils.typing import EllipsisType

#############################################################################
# Functions


def _inRange(
    x: np.array, rng: T.Sequence, lbi: bool = True, ubi: bool = False
) -> np.array:
    """`inRange` helper function.

    Parameters
    ----------
    x : array_like
        the array on which to test for membership in the range
        supports multiple dimension
    rng : list
        the range (lower, upper)
        when applied to ND array, must match the number of rows
        ex : for x [NxM], rng must be [Nx2]
    lbi : bool
        (default True)
        Lower Bound Inclusive, whether to be inclusive on the lower bound
    ubi : bool
        (default False)
        Upper Bound Inclusive, whether to be inclusive on the upper bound

    Returns
    -------
    idx: bool array
        bool index array
        shape matches `x`

    Examples
    --------
    >>> x = np.array([[ 0,  1],
    ...               [10, 11]])
    >>> rng = [[0, 3], [9, 11]]
    >>> _inRange(x, rng) # doctest: +SKIP
    array([[ True,  True],
           [ True, False]])

    """
    if len(x.shape) == 1:  # 1D

        if lbi and ubi:  # both true
            return (rng[0] <= x) & (x <= rng[1])
        elif lbi:  # only lbi is true
            return (rng[0] <= x) & (x < rng[1])
        elif ubi:  # only ubi is true
            return (rng[0] < x) & (x <= rng[1])
        else:  # neither true
            return (rng[0] < x) & (x < rng[1])

    else:  # ND

        # iterate over rows
        out = np.zeros_like(x, dtype=bool)
        for i, (row, lu) in enumerate(zip(x, rng)):
            out[i, :] = _inRange(row, lu, lbi=lbi, ubi=ubi)

        return out


# /def


# -----------------------------------------------------------------------------


@idxDecorator(_doc_style="numpy")
def inRange(
    *args: T.Union[np.array, T.Sequence],
    rng: T.Union[T.Sequence, EllipsisType] = Ellipsis,
    lbi: bool = True,
    ubi: bool = False
) -> np.array:
    """Multidimensional box selection.

    Parameters
    ----------
    args : Sequence
        list of values along each dimension
        must be the same length
        can be same-shaped ND arrays, each treated as a series of rows.
    rng : Sequence or Ellipsis, optional
        (default Ellipsis)
        the domain for each argument in `args`::

            args = [[x1], [x2], ...]
            rng =   [1st [lower, upper],
                     2nd [lower, upper],
                     ...]

        if each 'xn' is multidimensional
    lbi : bool, optional
        (default True)
        Lower Bound Inclusive, whether to be inclusive on the lower bound
    ubi : bool, optional
        (default False)
        Upper Bound Inclusive, whether to be inclusive on the upper bound
    as_ind : bool, optional
        (default False)
        whether to return bool array or the indices (where(bool array == True))
        sets the default behavior for the wrapped fnction *func*

    Returns
    -------
    inrange : bool ndarray
        boolean array to select values in box selection

    See Also
    --------
    outRange :  multidimensional box exclusion
    ioRange : `inRange` and `outRange` combined

    Examples
    --------
    list of args:

        >>> x = np.arange(5)
        >>> y = np.arange(5) + 10
        >>> inRange(x, y, rng=[[0, 3], [10, 15]]) # doctest: +SKIP
        array([ True,  True,  True, False, False])

    multidimensional arg:

        >>> x = np.array([[ 0,  1], [10, 11]])
        >>> rng = [[0, 3], [9, 11]]
        >>> inRange(x, rng=rng) # doctest: +SKIP
        array([[ True,  True],
               [ True, False]])

    .. todo::

        allow lbi & rbi to be lists, matching args, for individual adjustment

    """
    # if only one arg
    if len(args) == 1:
        rng = (rng,)

    rowbool = np.array(
        [_inRange(v, lu, lbi=lbi, ubi=ubi) for v, lu in zip(args, rng)]
    )

    numtrues = len(args)

    # now getting where all the dimensions are inside the bounds
    # collapses column to 1 row
    # check where all rows=True for each collapsed column
    allbool = rowbool.sum(axis=0)
    inrange = allbool == numtrues

    return inrange


# /def


# -----------------------------------------------------------------------------


@idxDecorator(_doc_style="numpy")
def outRange(
    *args: T.Union[np.array, T.Sequence],
    rng: T.Union[T.Sequence, EllipsisType] = Ellipsis,
    lbi: bool = True,
    ubi: bool = False
) -> np.array:
    """Multidimensional box exclusion.

    equivelent to ~inRange

    Parameters
    ----------
    args : Sequence
        either list of values along each dimension or list of values & bounds
        the input type depends on `rng`
    rng : Ellipsis or Sequence, optional
        (default Ellipsis)
        if rng is not Ellipsis::

            args = [[x1], [x2], ...]
            rng =   [1st [lower, upper],
                     2nd [lower, upper],
                     ...]

        else, args are the list of (x, [lower bound, upper. bound])
    lbi : bool, optional
        (default False)
        Lower Bound Inclusive, whether to be inclusive on the lower bound
    ubi : bool, optional
        (default True)
        Upper Bound Inclusive, whether to be inclusive on the upper bound
    as_ind : bool, optional
        (default False)
        whether to return bool array or the indices (where(bool array == True))
        sets the default behavior for the wrapped fnction *func*

    Returns
    -------
    outrange : ndarray of bool or ndarray of int
        boolean array to select values outside box selection
        if as_ind, then index array of same

    .. todo::

        allow `lbi` & `rbi` to be lists, matching `args`, for indiv adjustment

    """
    return ~inRange(*args, rng=rng, lbi=lbi, ubi=ubi)


# /def


# -----------------------------------------------------------------------------


@idxDecorator(_doc_style="numpy")
def ioRange(
    incl: T.Union[None, tuple, np.array] = None,
    excl: T.Union[None, tuple, np.array] = None,
    rng: T.Union[T.Sequence, EllipsisType] = Ellipsis,
) -> np.array:
    """Supports inRange and outRange.

    Parameters
    ----------
    incl : array_like, optional
        args into `inRange`
        must be a tuple if many args, not a tuple else
    excl : array_like, optional
        args into `outRange`
        must be a tuple if many args, not a tuple else
    rng : Sequence or Ellipsis, optional
        concatenated list of (in)outRange rng
        must be in orgder of [inRange rng, outRange rng]
    as_ind : bool
        (default False)
        whether to return bool array or the indices (``where(bool array == True``))
        sets the default behavior for the wrapped function

    Returns
    -------
    out : ndarray of bool or ndarray of ints
        boolean array to select values inside / outside box selection
        if as_ind, then index array of same

    """
    # Nothing passed. Raise error
    if (incl is None) & (excl is None):
        raise ValueError("incl and excl are None")
    # Only inclusion passed
    elif excl is None:
        if isinstance(incl, tuple):
            out = inRange(*incl, rng=rng)
        else:
            out = inRange(incl, rng=rng)
    # Only exclusion passed
    elif incl is None:
        if isinstance(excl, tuple):
            out = outRange(*excl, rng=rng)
        else:
            out = outRange(excl, rng=rng)
    # Both inclusion and exclusion
    else:
        if isinstance(incl, tuple):
            inclrng = rng[: len(incl)] if rng is not ... else ...
            out = inRange(*incl, rng=inclrng)
        else:
            inclrng = rng[: np.shape(incl)[0] - 1] if rng is not ... else ...
            if len(inclrng) == 1:
                inclrng = inclrng[0]
            out = inRange(incl, rng=inclrng)

        if isinstance(excl, tuple):
            exclrng = rng[len(excl) :] if rng is not ... else ...
            out &= outRange(*excl, rng=exclrng)
        else:
            exclrng = rng[np.shape(excl)[0] - 1 :] if rng is not ... else ...
            if len(exclrng) == 1:
                exclrng = exclrng[0]
            out &= outRange(excl, rng=exclrng)

    return out


# /def


# -----------------------------------------------------------------------------


@idxDecorator(_doc_style="numpy")
def ellipse(
    *x: T.Union[np.array, T.Sequence],
    x0: T.Union[float, T.Sequence] = 0.0,
    dx: T.Union[float, T.Sequence] = 1.0
) -> np.array:
    r"""Elliptical selection of data in many dimensions.

    Supports selection with variable center and radius::

        np.sum((x[i] - x0[i]) / dx[i])^2) < 1^2

    Parameters
    ----------
    x: m x (n, 1) array_like
        values along each dimension
    x0: scalar or (m, 1) array, optional
        (default = 0.)
        the center position of each x.
        can broadcast a scalar to apply to all
    dx: scalar or (m, 1) array, optional
        (default = 0.)
        the radius in each dimension

    Returns
    -------
    sel: array_like of bool
        bool array selecting data w/in ellipse
        if as_ind is True, then array_like of indices

    """
    shape = (len(x[0]), len(x))

    x0 = np.broadcast_to(x0, shape).T  # reshape x0 correctly
    dx = np.broadcast_to(dx, shape).T  # reshape x0 correctly

    arr = np.divide(np.subtract(x, x0), dx)

    return np.sqrt(np.sum(np.square(arr), axis=0)) < 1


# /def


# -----------------------------------------------------------------------------


@idxDecorator(_doc_style="numpy")
def circle(
    *x: T.Union[np.array, T.Sequence],
    x0: T.Union[float, T.Sequence] = 0.0,
    radius: T.Union[float, T.Sequence] = 1.0
) -> np.array:
    """Circular selection of data in many dimensions.

    Elliptical selection with fixed radius::

        np.sum((x[i] - x0[i]) / radius)^2) < 1^2

    where the summation is over the dimensions.

    Parameters
    ----------
    x : array_like
        shape (m x (n, 1))
        values along each dimension
    x0 : array_like, optional
        scalar or (m, 1) array
        the center position of each x, [default 0.]
        can broadcast a scalar to apply to all
    radius : float, optional
        the radius

    Returns
    -------
    sel : array_like of bool
        bool array selecting data within circle
        if as_ind is True, then array_like of indices

    """
    return ellipse(*x, x0=x0, dx=radius)


# /def

#############################################################################
# END



# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------
#
# TITLE   : astropy_imports
# PROJECT : utilipy
#
# ----------------------------------------------------------------------------

"""Base set of imports for :mod:`~astropy`.

Notes
-----
calls `quantity_support`

Routine Listings
----------------
Astropy: imports

    - units->u,
    - coordinates->coord, SkyCoord,
    - table.Table, QTable
    - visualization.quantity_support, astropy_mpl_style

References
----------
Astropy [#]_

.. [#] Astropy Collaboration et al., 2018, AJ, 156, 123.

"""

__author__ = "Nathaniel Starkman"


__all__ = [
    "astropy_imports_help",
    # "quantity_support",
    "astropy",
    "u",
    "coord",
    "SkyCoord",
    "Table",
    "QTable",
    "astropy_mpl_style",
    # modeling
    "modeling",
    "models",
    "fitting",
]


##############################################################################
# HELPER FUNCTIONS

# THIRD PARTY
import astropy
import astropy.coordinates as coord  # coordinates
import astropy.units as u  # units TODO replace with mine
from astropy import modeling
from astropy.coordinates import SkyCoord
from astropy.modeling import fitting, models
from astropy.table import QTable, Table  # table data structure
from astropy.visualization import (
    astropy_mpl_style,
    quantity_support,
    time_support,
)

# PROJECT-SPECIFIC
from utilipy.imports import conf
from utilipy.utils import make_help_function

##############################################################################
# IMPORTS


if float(astropy.__version__[:3]) > 4.1:
    from astropy.modeling import custom_model

    __all__ += ("custom_model",)


##############################################################################
# Cleaning Up

# astropy changes to matplotlib
quantity_support()
time_support()


##############################################################################
# Printing Information

astropy_imports_help = make_help_function(
    "astropy", __doc__, look_for="Routine Listings"
)


if conf.verbose_imports:
    astropy_imports_help()


##############################################################################
# END



# -*- coding: utf-8 -*-

"""Tests for `~utilipy.decorators.func_io`."""


__all__ = [
    "test_store_function_input",
    "test_add_folder_backslash",
    "test_random_generator_from_seed",
]


##############################################################################
# IMPORTS

# BUILT-IN
import functools
import inspect

# THIRD PARTY
import numpy as np
import pytest

# PROJECT-SPECIFIC
from utilipy.decorators import func_io

##############################################################################
# CODE
##############################################################################


def test_store_function_input():
    """Test :func:`~utilipy.decorators.func_io.store_function_input`."""
    # Inputs

    x, y = 1, 2

    # ----------
    # decorate by function call

    def test_func(x, y):
        """Test Function.

        Parameters
        ----------
        x, y : float

        """
        return x, y

    # /def

    # Store inputs
    store_func = func_io.store_function_input(
        test_func, store_inputs=True, _doc_style="numpy", _doc_fmt={}
    )
    res, inputs = store_func(x, y)  # pylint: disable=E1121

    assert res == (x, y)
    assert isinstance(inputs, inspect.BoundArguments)
    assert inputs.args == (x, y)

    # Don't store inputs
    no_store_func = func_io.store_function_input(
        test_func, store_inputs=False, _doc_style="numpy", _doc_fmt={}
    )
    res = no_store_func(x, y)  # pylint: disable=E1121

    assert res == (x, y)

    # ----------
    # decorate by pie-syntax

    @func_io.store_function_input  # all defaults
    def store_pie_func(x, y):
        """Test Function.

        Parameters
        ----------
        x, y : float

        """
        return x, y

    # /def

    res, inputs = store_pie_func(x, y)

    assert res == (x, y)
    assert isinstance(inputs, inspect.BoundArguments)
    assert inputs.args == (x, y)

    # ----------
    # decorate by called pie-syntax

    @func_io.store_function_input(store_inputs=True)  # just explicitly
    def store_pie_func2(x, y):
        """Test Function.

        Parameters
        ----------
        x, y : float

        """
        return x, y

    # /def

    res, inputs = store_pie_func2(x, y)

    assert res == (x, y)
    assert isinstance(inputs, inspect.BoundArguments)
    assert inputs.args == (x, y)


# /def


# -------------------------------------------------------------------


def test_add_folder_backslash():
    """Test :func:`~utilipy.decorators.func_io.add_folder_backslash`."""
    # ----------
    # decorate by function call

    def test_func(x, path):
        return path

    # /def

    args = (None, "~/Documents")

    # call as-is
    assert test_func(*args) == "~/Documents"

    # call decorated
    wrapped_func = func_io.add_folder_backslash(test_func, arguments=["path"])
    assert wrapped_func(*args) == "~/Documents/"  # pylint: disable=E1121

    # for str argument
    wrapped_func = func_io.add_folder_backslash(test_func, arguments="path")
    assert wrapped_func(*args) == "~/Documents/"  # pylint: disable=E1121

    # ----------
    # decorate by pie-syntax

    @func_io.add_folder_backslash(arguments=["path"])
    def wrapped_pie_func(x, path):
        return path

    assert wrapped_pie_func(None, "~/Documents") == "~/Documents/"
    assert wrapped_pie_func.__wrapped__(None, "~/Documents") == "~/Documents"


# /def


# -------------------------------------------------------------------


def _assert_LofLeq(list1, list2):
    """Assert nested lists are equal."""
    assert all(
        all(l1 == l2) if isinstance(l1, np.ndarray) else (l1 == l2)
        for l1, l2 in zip(list1, list2)
    )


def test_random_generator_from_seed():
    """Test :func:`~utilipy.decorators.func_io.random_generator_from_seed`."""
    exp = np.random.RandomState(0).get_state()
    # ----------
    # decorate by function call

    def test_func(x, random):
        return random

    # /def

    args = (None, 0)

    # call as-is
    assert test_func(*args) == 0

    # call decorated
    wrapped_fn = func_io.random_generator_from_seed(test_func)
    _assert_LofLeq(wrapped_fn(*args).get_state(), exp)  # pylint: disable=E1121

    # call wrapped partial
    wrapper = func_io.random_generator_from_seed()
    assert isinstance(wrapper, functools.partial)
    wrapped_fn = wrapper(test_func)
    _assert_LofLeq(wrapped_fn(*args).get_state(), exp)

    # call with str seed_names argument
    wrapped_fn = func_io.random_generator_from_seed(
        test_func, seed_names="random"
    )
    _assert_LofLeq(wrapped_fn(*args).get_state(), exp)  # pylint: disable=E1121

    # call with incorrect seed_names argument
    wrapped_fn = func_io.random_generator_from_seed(
        test_func, seed_names="not_here"
    )
    assert wrapped_fn(*args) == 0  # pylint: disable=E1124

    # call with strict input typing
    wrapped_fn = func_io.random_generator_from_seed(
        test_func, raise_if_not_int=True
    )
    with pytest.raises(TypeError):
        wrapped_fn(None, 0.1)  # pylint: disable=E1124

    # ----------
    # decorate by pie-syntax

    @func_io.random_generator_from_seed(seed_names="rng")
    def wrapped_pie_func(x, rng):
        return rng

    _assert_LofLeq(wrapped_pie_func(None, 0).get_state(), exp)


# /def

##############################################################################
# END



# -*- coding: utf-8 -*-
# see LICENSE.rst

"""Setup Script."""

__all__ = [
    "__version__",
]


##############################################################################
# CODE
##############################################################################

# this indicates whether or not we are in the package's setup.py
try:
    _ASTROPY_SETUP_  # type: ignore
except NameError:
    import builtins

    builtins._ASTROPY_SETUP_ = False

try:
    from .version import version as __version__
except ImportError:
    __version__ = ""


if not _ASTROPY_SETUP_:  # noqa
    import os
    from warnings import warn
    from astropy.config.configuration import (
        update_default_config,
        ConfigurationDefaultMissingError,
        ConfigurationDefaultMissingWarning,
    )

    # Create the test function for self test
    from astropy.tests.runner import TestRunner

    test = TestRunner.make_test_runner_in(os.path.dirname(__file__))
    test.__test__ = False
    __all__ += ["test"]

    # add these here so we only need to cleanup the namespace at the end
    config_dir = None

    if not os.environ.get("ASTROPY_SKIP_CONFIG_UPDATE", False):
        config_dir = os.path.dirname(__file__)
        config_template = os.path.join(config_dir, __package__ + ".cfg")
        if os.path.isfile(config_template):
            try:
                update_default_config(
                    __package__, config_dir, version=__version__
                )
            except TypeError as orig_error:
                try:
                    update_default_config(__package__, config_dir)
                except ConfigurationDefaultMissingError as e:
                    wmsg = (
                        e.args[0]
                        + " Cannot install default profile. If you are "
                        "importing from source, this is expected."
                    )
                    warn(ConfigurationDefaultMissingWarning(wmsg))
                    del e
                except Exception:
                    raise orig_error

# /if


##############################################################################
# END



# -*- coding: utf-8 -*-

"""Docstring inheritance-style implementations.

To implement your own inheritance file,
simply write a function to this template:

.. code-block:: python
    :linenos:

    def your_style(prnt_doc, child_doc, method='replace'):
        '''Merges parent and child docstrings

        Parameters
        ----------
        prnt_cls_doc: Optional[str]
        child_doc: Optional[str]
        method: Union[Literal["merge"], Literal["replace"]]

        Returns
        ------
        Optional[str]
            The merged docstring that will be utilized.

        '''
        return final_docstring


and log this using `custom_inherit.add_style(your_style)`. To permanently save
your function, define your function within custom_inherit/_style_store.py, and
log it in custom_inherit.style_store.__all__. Your style will then be
available as 'your_style' (i.e. whatever you named the function).

Routine Listings
----------------
parent
numpy
reST
google
numpy_napoleon

References
----------
The code source [1]_.

.. [1] https://github.com/rsokl/custom_inherit


"""

__author__ = "Ryan Soklaski"
__license__ = "MIT"
__version__ = "2.2.2"
__email__ = "https://github.com/rsokl/custom_inherit"

# All built-in styles must be logged in the __all__ field.
__all__ = ["parent", "numpy", "reST", "google", "numpy_napoleon"]


###############################################################################
# IMPORTS

# BUILT-IN
import typing as T

# THIRD PARTY
from typing_extensions import Literal

# PROJECT-SPECIFIC
from .napoleon_parse_tools import (
    merge_google_napoleon_docs,
    merge_numpy_napoleon_docs,
)
from .numpy_parse_tools import merge_numpy_docs
from .rest_parse_tools import merge_rest_docs

###############################################################################
# CODE
###############################################################################


def parent(
    prnt_doc: T.Union[str, None],
    child_doc: T.Union[str, None],
    method="replace",
):
    """Documentation strings are inherited if not overridden.

    Wherever the docstring for a child-class' attribute (or for the class
    itself) is `None`, inherit the corresponding docstring from the parent.

    Parameters
    ----------
    prnt_sec: Optional[str]
        The docstring section from the parent's attribute.
    child_sec: Optional[str]
        The docstring section from the child's attribute.
    method: {replace}, optional
        'replace' (only option) parent with child sections, if exists.

    *NOTE* As of Python 3.5, this is the default behavior of the built-in
    function inspect.getdoc

    """
    return child_doc if child_doc is not None else prnt_doc


# /def


def numpy(
    prnt_doc: T.Union[str, None],
    child_doc: T.Union[str, None],
    method: T.Union[Literal["merge"], Literal["replace"]] = "replace",
):
    """Merges numpy-styled docstrings from the parent and child.

    Specifically, any docstring section that appears in the parent's docstring
    that is not present in the child's is inherited. Otherwise, the child's
    docstring section is utilized. An exception to this is if the parent
    docstring contains a "Raises" section, but the child's attribute's
    docstring contains a "Returns" or "Yields" section instead. In this
    instance, the "Raises" section will not appear in the inherited docstring.

    Any whitespace that can be uniformly removed from a docstring's second line
    and onwards is removed. Sections in the resulting docstring will be
    separated by a single blank line.

    For details on the numpy docstring style, see:
    https://github.com/numpy/numpy/blob/master/doc/HOWTO_DOCUMENT.rst.txt

    Parameters
    ----------
    prnt_sec: Optional[str]
        The docstring section from the parent's attribute.
    child_sec: Optional[str]
        The docstring section from the child's attribute.
    method: {merge, replace}, optional
        'merge' merges the sections.
        'replace' (default) parent with child sections, if exists.

    Examples
    --------
    parent's docstring:

        ''' Parent's line

            Parameters
            ----------
            x: int
                description of x
            y: Union[None, int]
                description of y

            Raises
            ------
            NotImplemented Error

        '''

    child's docstring:

        ''' Child's line

            Returns
            -------
            int

            Notes
            -----
            notes blah blah

        '''

    docstring that is ultimately inherited:

        ''' Child's line

            Parameters
            ----------
            x: int
                description of x
            y: Union[None, int]
                description of y

            Returns
            -------
            int

            Notes
            -----
            notes blah blah

        '''

    """
    return merge_numpy_docs(prnt_doc, child_doc, method=method)


# /def


def reST(
    prnt_doc: T.Union[str, None],
    child_doc: T.Union[str, None],
    method: T.Union[Literal["merge"], Literal["replace"]] = "replace",
):
    """Merge two reST-style docstrings into a single docstring.

    Given the reST-style docstrings from a parent and child's attributes,
    merge the docstring sections such that the child's section is used,
    wherever present, otherwise the parent's section is used.

    Sections are delimited by any type of reST section title. For more
    details, see:
    http://docutils.sourceforge.net/docs/ref/rst/restructuredtext.html#sections

    Any whitespace that can be uniformly removed from a docstring's second
    line and onwards is removed. Sections in the resulting docstring will be
    separated by a single blank line.

    Parameters
    ----------
    prnt_sec: Optional[str]
        The docstring section from the parent's attribute.
    child_sec: Optional[str]
        The docstring section from the child's attribute.
    method: {merge, replace}, optional
        'merge' merges the sections.
        'replace' (default) parent with child sections, if exists.

    Examples
    --------
    parent:

        '''
        Header1
        -------
        parent's content for Header 1
           - indented material

        Header2
        ~~~~~~~
        content for Header2

       '''

    child:

        '''Front-matter

        ~~~~~~~~~
        NewHeader
        ~~~~~~~~~
        content for NewHeader

        Header2
        +++++++
        child's content for Header2

       '''

    merged:

        '''Front-matter

        Header1
        -------
        content for Header 1
           - indented material

        Header2
        +++++++
        child's content for Header2

        ~~~~~~~~~
        NewHeader
        ~~~~~~~~~
        content for NewHeader

       '''

    """
    return merge_rest_docs(prnt_doc, child_doc, method=method)


# /def


def numpy_napoleon(
    prnt_doc: T.Union[str, None],
    child_doc: T.Union[str, None],
    method: T.Union[Literal["merge"], Literal["replace"]] = "replace",
):
    """Behaves identically to the 'numpy' style, but abides by the docstring sections
    specified by the "Napoleon" standard.

    For more info regarding the Napoleon standard, see:
    http://sphinxcontrib-napoleon.readthedocs.io/en/latest/index.html#docstring-sections

    Parameters
    ----------
    prnt_sec: Optional[str]
        The docstring section from the parent's attribute.
    child_sec: Optional[str]
        The docstring section from the child's attribute.
    method: {merge, replace}, optional
        'merge' merges the sections.
        'replace' (default) parent with child sections, if exists.

    Examples
    --------
    parent's docstring:

        '''Parent's line

        Keyword Arguments
        -----------------
        x: int
            description of x
        y: Union[None, int]
            description of y

        Raises
        ------
        NotImplemented Error

        '''

    child's docstring:

        '''Child's line

        Returns
        -------
        int

        Notes
        -----
        notes blah blah

        '''

    docstring that is ultimately inherited:

        '''Child's line

        Keyword Arguments
        -----------------
        x: int
            description of x
        y: Union[None, int]
            description of y

        Returns
        -------
        int

        Notes
        -----
        notes blah blah

        '''

    """
    return merge_numpy_napoleon_docs(prnt_doc, child_doc, method=method)


# /def


def google(
    prnt_doc: T.Union[str, None],
    child_doc: T.Union[str, None],
    method: T.Union[Literal["merge"], Literal["replace"]] = "replace",
):
    """Merge google-styled docstrings by the "Napoleon" standard.

    Specifically, any docstring section that appears in the parent's docstring
    that is not present in the child's is inherited. Otherwise, the child's
    docstring section is utilized. An exception to this is if the parent
    docstring contains a "Raises" section, but the child's attribute's
    docstring contains a "Returns" or "Yields" section instead. In this
    instance, the "Raises" section will not appear in the inherited docstring.

    Any whitespace that can be uniformly removed from a docstring's second line
    and onwards is removed. Sections in the resulting docstring will be
    separated by a single blank line.

    For more info regarding the Napoleon standard, see:
    http://sphinxcontrib-napoleon.readthedocs.io/en/latest/index.html#docstring-sections

    For details on the google docstring style, see:
    http://sphinxcontrib-napoleon.readthedocs.io/en/latest/example_google.html#example-google

    Parameters
    ----------
    prnt_sec: Optional[str]
        The docstring section from the parent's attribute.
    child_sec: Optional[str]
        The docstring section from the child's attribute.
    method: {merge, replace}, optional
        'merge' merges the sections.
        'replace' (default) parent with child sections, if exists.

    Examples
    --------
    parent's docstring:

        '''Parent's line

        Args:
            x: int
                description of x
            y: Union[None, int]
                description of y

        Raises:
            NotImplemented Error

        '''

    child's docstring:

        '''Child's line

        Returns:
            int

        Notes:
            notes blah blah

        '''

    docstring that is ultimately inherited:

        '''Child's line

        Parameters:
            x: int
                description of x
            y: Union[None, int]
                description of y

        Returns:
            int

        Notes:
            notes blah blah

        '''

    """
    return merge_google_napoleon_docs(prnt_doc, child_doc, method=method)


# /def

##############################################################################
# END



# -*- coding: utf-8 -*-
"""Configure Test Suite.

This file is used to configure the behavior of pytest when using the Astropy
test infrastructure. It needs to live inside the package in order for it to
get picked up when running the tests inside an interpreter using
packagename.test

"""

##############################################################################
# IMPORTS

# BUILT-IN
import os

# THIRD PARTY
import pytest
from astropy.version import version as astropy_version

# For Astropy 3.0 and later, we can use the standalone pytest plugin
if astropy_version < "3.0":
    from astropy.tests.pytest_plugins import *  # noqa

    del pytest_report_header
    ASTROPY_HEADER = True

else:
    try:
        from pytest_astropy_header.display import (
            PYTEST_HEADER_MODULES,
            TESTED_VERSIONS,
        )

        ASTROPY_HEADER = True
    except ImportError:
        ASTROPY_HEADER = False

# /if

##############################################################################
# CODE
##############################################################################


def pytest_configure(config):
    """Configure Pytest with Astropy.

    Parameters
    ----------
    config : pytest configuration

    """
    if ASTROPY_HEADER:

        config.option.astropy_header = True

        # Customize the following lines to add/remove entries from the list of
        # packages for which version numbers are displayed when running the
        # tests.
        PYTEST_HEADER_MODULES.pop("Pandas", None)
        PYTEST_HEADER_MODULES["scikit-image"] = "skimage"

        from . import __version__

        packagename = os.path.basename(os.path.dirname(__file__))
        TESTED_VERSIONS[packagename] = __version__


# /def

# ------------------------------------------------------


@pytest.fixture(autouse=True)
def add_astropy(doctest_namespace):
    """Add Astropy stuff to Pytest.

    Parameters
    ----------
    doctest_namespace : namespace

    """
    import astropy.units

    # add to namespace
    doctest_namespace["u"] = astropy.units

    return


# def


# ------------------------------------------------------


# Uncomment the last two lines in this block to treat all DeprecationWarnings
# as exceptions. For Astropy v2.0 or later, there are 2 additional keywords,
# as follow (although default should work for most cases).
# To ignore some packages that produce deprecation warnings on import
# (in addition to 'compiler', 'scipy', 'pygments', 'ipykernel', and
# 'setuptools'), add:
#     modules_to_ignore_on_import=['module_1', 'module_2']
# To ignore some specific deprecation warning messages for Python version
# MAJOR.MINOR or later, add:
#     warnings_to_ignore_by_pyver={(MAJOR, MINOR): ['Message to ignore']}
# from astropy.tests.helper import enable_deprecations_as_exceptions  # noqa
# enable_deprecations_as_exceptions()

##############################################################################
# END



# -*- coding: utf-8 -*-

"""Test subpackages for :mod:`~utilipy.data_utils`."""


__all__ = [
    "test_init_fitting",
    "test_init_xfm",
]


##############################################################################
# IMPORTS


# PROJECT-SPECIFIC
from utilipy.data_utils import fitting, xfm

##############################################################################
# PARAMETERS


##############################################################################
# CODE
##############################################################################


def test_init_fitting():
    """Test module initialization."""
    # Expectations
    local = [
        # modules
        "lmfit_utils",
        # functions
        "scipy_residual_to_lmfit",
    ]

    # test __all__ conforms to module
    for name in fitting.__all__:
        assert hasattr(fitting, name)

    # test __all__ matches expectations
    for name in fitting.__all__:
        assert name in local

    return


# /def

# -------------------------------------------------------------------


def test_init_xfm():
    """Test module initialization."""
    # Expectations
    local = [
        # modules
        "graph",
        "transformations",
        # functions
        "TransformGraph",
        # transformations
        "DataTransform",
        "CompositeTransform",
        # Parameters
        "data_graph",
    ]

    # test __all__ conforms to module
    for name in xfm.__all__:
        assert hasattr(xfm, name)

    # test __all__ matches expectations
    for name in xfm.__all__:
        assert name in local

    return


# /def


# def test_fitting_top_level_imports():
#     """Test top-level imports of :mod:`~utilipy.data_utils.fitting`."""
#     # First test they exist
#     subpkg: str
#     for subpkg in fitting.__all_top_imports__:
#         assert hasattr(fitting, subpkg)

#     # Next test that top-levels are all the possible top-levels
#     drct: str = os.path.split(fitting.__file__)[0]  # directory
#     donottest = ("__pycache__", "data", "tests")  # stuff not to test

#     for file in os.listdir(drct):  # iterate through directory
#         # test?
#         if os.path.isdir(drct + "/" + file) and file not in donottest:
#             assert file in fitting.__all_top_imports__
#         else:  # nope, chuck testa.
#             pass

#     return


# # /def


# -------------------------------------------------------------------


##############################################################################
# END



# -*- coding: utf-8 -*-

"""IPython Imports.

Routine Listings
----------------
run_imports
    Import file, custom or provided, using ipython magic
import_from_file
    Run import(s) from a file(s)
aimport
    from `.autoreload` module
set_autoreload
    from `.autoreload` module

import_base
    ipython magic run `utilipy/imports/base.py`
import_extended
    ipython magic run `utilipy/imports/extended.py`
import_astropy
    ipython magic run `utilipy/imports/astropy.py`
import_matplotlib
    ipython magic run `utilipy/imports/matplotlib.py`
import_galpy
    ipython magic run `utilipy/imports/galpy.py`
import_amuse
    ipython magic run `utilipy/imports/amuse.py`

References
----------
IPython [#]_

.. [#] Fernando Pérez, Brian E. Granger, IPython: A System for Interactive
    Scientific Computing, Computing in Science and Engineering, vol. 9,
    no. 3, pp. 21-29, May/June 2007, doi:10.1109/MCSE.2007.53.
    URL: https://ipython.org

"""

__author__ = "Nathaniel Starkman"


__all__ = [
    "import_from_file",
    "run_imports",
    "aimport",
    "set_autoreload",
    # specific importers
    "import_base",
    "import_extended",
    "import_astropy",
    "import_matplotlib",  # 'import_plotly',
    "import_galpy",
    "import_amuse",
    "import_astronat",
]


##############################################################################
# IMPORTS

# BUILT-IN
import typing as T
from pathlib import Path

# THIRD PARTY
from IPython import get_ipython

# PROJECT-SPECIFIC
from .autoreload import aimport, set_autoreload
from utilipy.data_utils import get_path_to_file
from utilipy.decorators.docstring import set_docstring_for_import_func
from utilipy.imports import use_import_verbosity
from utilipy.utils.logging import LogFile

##############################################################################
# SETUP

_LOGFILE = LogFile(header=False, verbose=0)
_LOGGER_KW = {"print": False}


##############################################################################
# CODE
##############################################################################


def import_from_file(
    *files: str,
    is_relative: bool = True,
    verbose_imports: T.Optional[bool] = None,
    logger: LogFile = _LOGFILE,
    verbose=None,
    logger_kw: T.Optional[T.Dict[str, T.Any]] = _LOGGER_KW,
) -> None:
    """Run import(s) from a file(s).

    Parameters
    ----------
    *files: str(s)
        strings for files to import
        need to include file suffix
    is_relative: bool or list of bools
        whether the `*files` paths are relative or absolute
    verbose_imports: bool or None
        Verbose_imports or not, use default if None.

    """
    # Handle `relative`
    if isinstance(is_relative, bool):  # broadcasting to same length as files
        relative = [is_relative] * len(files)
    else:  # already a list
        if len(is_relative) != len(files):  # checking correct length
            raise ValueError("len(relative) != len(files)")
        relative = is_relative

    with use_import_verbosity(verbose_imports):  # handle verbose-imports

        for file, rltv in zip(files, relative):  # loop over imports
            if rltv:
                file = str(Path(file).resolve())

            # get_ipython inbuilt to jupyter
            get_ipython().magic(f"run {file}")

            # logging
            # implemented separately b/c files often have own print statements
            logger_kw = logger_kw or {}
            logger.report(f"imported {file}", verbose=verbose, **logger_kw)

    return


# /def


# ----------------------------------------------------------------------------


def run_imports(
    *files: str,
    is_relative: bool = True,
    # standard import files
    base: bool = False,
    extended: bool = False,
    # extra standard files
    astropy: bool = False,
    matplotlib: bool = False,
    plotly: bool = False,
    # additional, requires extra installs
    galpy: bool = False,
    amuse: bool = False,
    astronat: bool = False,
    # utils
    verbose_imports: T.Optional[bool] = None,
    # logging
    logger: LogFile = _LOGFILE,
    verbose: T.Optional[int] = 0,
    logger_kw: T.Optional[T.Dict[str, T.Any]] = None,
) -> None:
    """Import file using IPython magic.

    if `astropy` and `matplotlib`, sets matplotlib style to astropy_mpl_style

    Parameters
    ----------
    files: str(s)
        strings for files to import
        need to include file suffix
    base: bool
        a broad set of basic imports
        import_base -> `utilipy/imports/base_imports.py`
    astropy: bool
        import_astropy -> `utilipy/imports/astropy_imports.py`
    matplotlib: bool
        import_matplotlib -> `utilipy/imports/matplotlib_imports.py`
    plotly: bool
        import_plotly -> `utilipy/imports/plotly_imports.py`
    extended: bool
        import_extended -> `utilipy/imports/extended_imports.py`
    galpy: bool
        import_galpy -> `utilipy/imports/galpy_imports.py`
    amuse: bool
        import_amuse -> `utilipy/imports/amuse_imports.py`
    astronat : bool
        import_astronat -> `utilipy/imports/astronat_imports.py`

    Other Parameters
    ----------------
    relative: bool or list of bools
        whether the `files` paths are relative or absolute
    verbose_imports: bool or None
        Verbose_imports or not, use ``.utilipyrc`` default if None.

    Examples
    --------
    ``run_imports(base=True, astropy=True)``
    imports from `utilipy/imports/base.py` and `utilipy/imports/astropy.py`,
    printing an import summary

    ``run_imports(base=True, verbose_imports=False)``
    imports from `utilipy/imports/base.py`, without an import summary

    ``utilipy.imports.conf.verbose_imports = False``
    ``utilipy.ipython.imports.run_imports(base=True, verbose_imports=None)``
    imports from `utilipy/imports/base.py` with default import-verbosity state

    """
    # make kwargs that go into every standard import
    kw = dict(verbose_imports=verbose_imports, logger=logger, verbose=verbose)
    logger_kw = logger_kw or {}

    if verbose_imports:
        print("Importing:")

    # ---------------------------------------------
    # base

    if base:
        import_base(**kw)

    if extended:
        import_extended(**kw)

    # ---------------------------------------------
    # basic

    if astropy:
        import_astropy(**kw)

    if matplotlib:
        import_matplotlib(**kw)
    if plotly:
        import_plotly(**kw)

    # ---------------------------------------------
    # extras

    if galpy:
        import_galpy(**kw)

    if amuse:
        import_amuse(**kw)

    if astronat:
        import_astronat(**kw)

    # ---------------------------------------------

    # when combined
    if astropy & matplotlib:
        from matplotlib import pyplot
        from astropy.visualization import astropy_mpl_style

        pyplot.style.use(astropy_mpl_style)

    # TODO, embed in galpy_imports using argparse
    if galpy & amuse:
        from galpy.potential import to_amuse  # noqa

    # other import filess
    if files:  # True if not empty
        import_from_file(
            *files,
            is_relative=is_relative,
            verbose_imports=verbose_imports,
            logger=logger,
            verbose=verbose,
            logger_kw=logger_kw,
        )

    return


# /def


##############################################################################
# Specific Imports
# TODO make these with a function


@set_docstring_for_import_func(
    "base_imports.py", package="utilipy.imports", section="Returns"
)
def import_base(
    verbose_imports: T.Optional[bool] = None,
    logger: LogFile = _LOGFILE,
    verbose: T.Optional[int] = None,
    logger_kw: T.Optional[T.Dict[str, T.Any]] = _LOGGER_KW,
) -> None:
    """Import base packages."""
    import_from_file(
        get_path_to_file("base_imports.py", package="utilipy.imports"),
        is_relative=False,
        verbose_imports=verbose_imports,
        logger=logger,
        verbose=verbose,
        logger_kw=logger_kw,
    )
    return


# /def


@set_docstring_for_import_func(
    "extended_imports.py",
    package="utilipy.imports",
    section="Routine Listings",
)
def import_extended(
    verbose_imports: T.Optional[bool] = None,
    logger: LogFile = _LOGFILE,
    verbose: T.Optional[int] = None,
    logger_kw: T.Optional[T.Dict[str, T.Any]] = _LOGGER_KW,
) -> None:
    """Import extended packages."""
    import_from_file(
        get_path_to_file("extended_imports.py", package="utilipy.imports"),
        is_relative=False,
        verbose_imports=verbose_imports,
        logger=logger,
        verbose=verbose,
        logger_kw=logger_kw,
    )
    return


# /def


@set_docstring_for_import_func(
    "astropy_imports.py", package="utilipy.imports", section="Routine Listings"
)
def import_astropy(
    verbose_imports: T.Optional[bool] = None,
    logger: LogFile = _LOGFILE,
    verbose: T.Optional[int] = None,
    logger_kw: T.Optional[T.Dict[str, T.Any]] = _LOGGER_KW,
) -> None:
    """Import basic astropy packages."""
    import_from_file(
        get_path_to_file("astropy_imports.py", package="utilipy.imports"),
        is_relative=False,
        verbose_imports=verbose_imports,
        logger=logger,
        verbose=verbose,
        logger_kw=logger_kw,
    )
    return


# /def


# ----------------------------------------------------------------------------
# plotting


@set_docstring_for_import_func(
    "matplotlib_imports.py",
    package="utilipy.imports",
    section="Routine Listings",
)
def import_matplotlib(
    verbose_imports: T.Optional[bool] = None,
    logger: LogFile = _LOGFILE,
    verbose: T.Optional[int] = None,
    logger_kw: T.Optional[T.Dict[str, T.Any]] = _LOGGER_KW,
) -> None:
    """Import basic Matplotlib packages."""
    import_from_file(
        get_path_to_file("matplotlib_imports.py", package="utilipy.imports"),
        is_relative=False,
        verbose_imports=verbose_imports,
        logger=logger,
        verbose=verbose,
        logger_kw=logger_kw,
    )
    return


# /def


@set_docstring_for_import_func(
    "plotly_imports.py", package="utilipy.imports", section="Routine Listings"
)
def import_plotly(
    verbose_imports: T.Optional[bool] = None,
    logger: LogFile = _LOGFILE,
    verbose: T.Optional[int] = None,
    logger_kw: T.Optional[T.Dict[str, T.Any]] = _LOGGER_KW,
) -> None:
    """Import basic Matplotlib packages."""
    import_from_file(
        get_path_to_file("plotly_imports.py", package="utilipy.imports"),
        is_relative=False,
        verbose_imports=verbose_imports,
        logger=logger,
        verbose=verbose,
        logger_kw=logger_kw,
    )
    return


# /def


# ----------------------------------------------------------------------------
# extras


@set_docstring_for_import_func(
    "galpy_imports.py", package="utilipy.imports", section="Routine Listings"
)
def import_galpy(
    verbose_imports: T.Optional[bool] = None,
    logger: LogFile = _LOGFILE,
    verbose: T.Optional[int] = None,
    logger_kw: T.Optional[T.Dict[str, T.Any]] = _LOGGER_KW,
) -> None:
    """Import basic galpy packages."""
    import_from_file(
        get_path_to_file("galpy_imports.py", package="utilipy.imports"),
        is_relative=False,
        verbose_imports=verbose_imports,
        logger=logger,
        verbose=verbose,
        logger_kw=logger_kw,
    )
    return


# /def


@set_docstring_for_import_func(
    "amuse_imports.py", package="utilipy.imports", section="Routine Listings"
)
def import_amuse(
    verbose_imports: T.Optional[bool] = None,
    logger: LogFile = _LOGFILE,
    verbose: T.Optional[int] = None,
    logger_kw: T.Optional[T.Dict[str, T.Any]] = _LOGGER_KW,
) -> None:
    """Import basic amuse packages."""
    import_from_file(
        get_path_to_file("amuse_imports.py", package="utilipy.imports"),
        is_relative=False,
        verbose_imports=verbose_imports,
        logger=logger,
        verbose=verbose,
        logger_kw=logger_kw,
    )
    return


# /def


@set_docstring_for_import_func(
    "astronat_imports.py",
    package="utilipy.imports",
    section="Routine Listings",
)
def import_astronat(
    verbose_imports: T.Optional[bool] = None,
    logger: LogFile = _LOGFILE,
    verbose: T.Optional[int] = None,
    logger_kw: T.Optional[T.Dict[str, T.Any]] = _LOGGER_KW,
) -> None:
    """Import basic amuse packages."""
    import_from_file(
        get_path_to_file("astronat_imports.py", package="utilipy.imports"),
        is_relative=False,
        verbose_imports=verbose_imports,
        logger=logger,
        verbose=verbose,
        logger_kw=logger_kw,
    )
    return


# /def


##############################################################################
# END



# -*- coding: utf-8 -*-

"""Math Core Functions."""

__author__ = "Nathaniel Starkman"


__all__ = [
    "quadrature",
    # quantity functions
    "as_quantity",
    "qsquare",
    "qnorm",
    "qarange",
]


##############################################################################
# IMPORTS

# BUILT-IN
import typing as T
import warnings

# THIRD PARTY
import numpy as np
from astropy.units import Quantity
from numpy.linalg import norm

##############################################################################
# IMPORTS

_NUMPY_VERSION = [
    int(v) for i, v in enumerate(np.__version__.split(".")) if i < 3
]
if (_NUMPY_VERSION[0] <= 1) and (_NUMPY_VERSION[1] <= 16):  # v1.16
    warnings.warn(
        (
            "Need to set the environment variable "
            "NUMPY_EXPERIMENTAL_ARRAY_FUNCTION=1"
        )
    )


###############################################################################
# CODE
###############################################################################


def quadrature(*args: T.Tuple[T.Sequence], axis: int = 0) -> T.Sequence:
    """Return arguments summed in quadrature.

    ::

        sqrt(sum(square(args), axis=axis))

    Parameters
    ----------
    args: Sequence
    axis: int
        the summation axis, default 0.

    Returns
    -------
    array-like

    """
    if len(args) == 0:
        raise ValueError
    if len(args) == 1:
        args = args[0]

    return np.sqrt(np.sum(np.square(args), axis=axis))


# /def


#####################################################################
# Quantity Functions


def as_quantity(arg):
    """Convert argument to a Quantity (or raise NotImplementedError).

    from :mod:`~astropy.utils`.

    Returns
    -------
    Quantity
        not copied, quantity subclasses passed through.

    Raises
    ------
    NotImplementedError
        if Quantity() fails

    """
    try:
        return Quantity(arg, copy=False, subok=True)
    except Exception:
        raise NotImplementedError


# /def

# -------------------------------------------------------------------


def qsquare(*args, **kw):
    """Quantity, Squared.

    Parameters
    ----------
    *args : Quantity
        passed, as tuple, to :func:`~as_quantity`
    **kw
        arguments into :func:`~numpy.square`

    Returns
    -------
    Quantity
        not copied, quantity subclasses passed through.

    Raises
    ------
    NotImplementedError
        if :func:`~as_quantity` fails

    """
    return np.square(as_quantity(args), **kw)


# /def

# -------------------------------------------------------------------


def qnorm(*args, **kw):
    """Quantity, Normed.

    Parameters
    ----------
    *args : Quantity
        passed, as tuple, to :func:`~as_quantity`
    **kw
        arguments into :func:`~numpy.linalg.norm`

    Returns
    -------
    Quantity
        not copied, quantity subclasses passed through.

    Raises
    ------
    NotImplementedError
        if :func:`~as_quantity` fails

    """
    return norm(as_quantity(args), **kw)


# /def


# -------------------------------------------------------------------


def qarange(start, stop, step, unit=None):
    """:func:`~numpy.arange` for Quantities."""
    if unit is None:
        unit = step.unit

    arng = np.arange(
        start.to_value(unit), stop.to_value(unit), step.to_value(unit)
    )

    return arng * unit


# /def


# -------------------------------------------------------------------


###############################################################################
# END



# -*- coding: utf-8 -*-

"""Tests for utilipy initialization."""


##############################################################################
# IMPORTS

# BUILT-IN
import os

# THIRD PARTY
import pytest

# PROJECT-SPECIFIC
from utilipy import __all_top_imports__ as all_top_imports
from utilipy import __dict__ as D
from utilipy import __file__ as _file
from utilipy import imports, lookup, online_help, reload_config

##############################################################################
# TESTS
##############################################################################


def test_reload_config():
    """Test :func:`~utilipy.reload_config`."""
    # get initial value
    init_value = imports.conf.verbose_imports

    # set changed value and test
    imports.conf.verbose_imports = bool(~init_value)
    assert imports.conf.verbose_imports == bool(~init_value)

    # reload and test
    reload_config()
    assert imports.conf.verbose_imports == init_value


# /def


# --------------------------------------------------------------------------


@pytest.mark.remote_data
def test_online_help():  # TODO actual tests.
    """Test :fun:`~utilipy.online_help`."""
    # -----------
    # by object
    online_help(query=online_help)
    online_help(query=online_help, version="1.0")

    # -----------
    # by str
    online_help(query="online_help")
    online_help(query="online_help", version="1.0")

    # -----------
    # blank
    online_help(query=None)
    online_help(query=None, version="1.0")


# /def


# --------------------------------------------------------------------------


@pytest.mark.remote_data
def test_lookup():  # TODO actual tests.
    """Test :fun:`~utilipy.online_help`."""
    lookup(query="lookup", online=True)

    lookup(query="lookup", online=False)


# /def


# --------------------------------------------------------------------------


def test_top_level_imports():
    """Test Top-Level Imports."""
    # First test they exist
    subpkg: str
    for subpkg in all_top_imports:
        assert subpkg in D

    # Next test that top-levels are all the possible top-levels
    drct: str = os.path.split(_file)[0]  # directory
    donottest = ("tests", "__pycache__", "ipython")  # stuff not to test

    for file in os.listdir(drct):  # iterate through directory
        if os.path.isdir(drct + "/" + file) and file not in donottest:
            assert file in all_top_imports
        # else:  # nope, chuck testa.
        #     pass


# /def


# --------------------------------------------------------------------------


def test_specific_imports():
    """Test specific imports."""
    # imports same as __init__
    import utilipy

    from utilipy.utils.logging import LogFile
    from utilipy.utils.collections import ObjDict
    from utilipy.utils.functools import wraps

    # test equality
    assert utilipy.LogFile is LogFile
    assert utilipy.ObjDict is ObjDict
    assert utilipy.wraps is wraps


# /def


##############################################################################
# END



# -*- coding: utf-8 -*-

"""Functions for working with autoreload extension.

Routine Listings
----------------
`aimport`
    Jupyter magic aimport.

`set_autoreload`
    Global imports setting.

References
----------
IPython [1]_


.. [1] Fernando Pérez, Brian E. Granger, IPython: A System for Interactive
    Scientific Computing, Computing in Science and Engineering, vol. 9,
    no. 3, pp. 21-29, May/June 2007, doi:10.1109/MCSE.2007.53.
    URL: https://ipython.org

"""


__all__ = [
    "set_autoreload",
    "aimport",
]


##############################################################################
# IMPORTS

# BUILT-IN
import typing as T

# THIRD PARTY
from IPython import get_ipython

# PROJECT-SPECIFIC
from utilipy.utils.logging import LogPrint

##############################################################################
# PARAMETERS

_LOGFILE = LogPrint(header=False, verbose=0)

try:
    get_ipython()
    if get_ipython() is None:  # double checking
        raise NameError
except NameError:
    _HAS_IPY: bool = False
else:
    _HAS_IPY: bool = True


##############################################################################
# CODE
##############################################################################


def set_autoreload(reload_type: T.Optional[int] = None):
    """Global imports setting.

    Parameters
    ----------
    reload_type: {0, 1, 2} or None
        (default None)

        - **0**) Disable automatic reloading.
        - **1**) Reload all modules imported with %aimport
                 before executing the Python code typed.
        - **2**) Reload all modules (except those excluded by %aimport)
                 before executing the Python code typed.
        - **None**) null. do not change current setting.

    See Also
    --------
    autoreload: module
        `%load_ext autoreload`
        `%aimport`

    """
    if reload_type is not None:

        # set autoreload type
        get_ipython().magic(f"autoreload {reload_type}")

        _LOGFILE.write(f"set autoreload to {reload_type}")

    return


# /def


##############################################################################


def aimport(*modules: str, autoreload: T.Union[bool, list, tuple] = True):
    """Jupyter magic aimport.

    Parameters
    ----------
    modules: list of strs
        the modules to be imported
    autoreload: bool or list, optional
        (default True)
        whether the imported modules are marked for autoreloading
        if its a list, it must be the same length as `modules`

    """
    # making autoreload compatible with modules
    if isinstance(autoreload, bool):
        autoreload = [autoreload] * len(modules)
    elif len(autoreload) == len(modules):  # any list
        pass
    else:
        raise ValueError("len(autoreload) != len(modules)")

    for module, reload_type in zip(modules, autoreload):
        # testing correct data types
        assert isinstance(module, str)
        assert isinstance(reload_type, bool)

        if not reload_type:
            module = "-" + module  # mark for not autoreloading

        # import
        get_ipython().magic(f"aimport {module}")

    return


# /def


##############################################################################
# SETTING STATE

if _HAS_IPY is True:
    get_ipython().magic("load_ext autoreload")  # autoreload extensions
    set_autoreload(1)
# /if

##############################################################################
# END



# -*- coding: utf-8 -*-

"""Metaclasses."""

__author__ = "Nathaniel Starkman"
__credits__ = ["astropy"]


##############################################################################
# IMPORTS

# BUILT-IN
import inspect
import typing as T

##############################################################################
# CODE


class InheritDocstrings(type):
    """Docstring inheritance metaclass.

    This metaclass makes methods of a class automatically have their
    docstrings filled in from the methods they override in the base
    class.
    If the class uses multiple inheritance, the docstring will be
    chosen from the first class in the bases list, in the same way as
    methods are normally resolved in Python.  If this results in
    selecting the wrong docstring, the docstring will need to be
    explicitly included on the method.

    For example::
        >>> class A(metaclass=InheritDocstrings):
        ...     def wiggle(self):
        ...         "Wiggle the thingamajig"
        ...         pass
        >>> class B(A):
        ...     def wiggle(self):
        ...         pass
        >>> B.wiggle.__doc__
        'Wiggle the thingamajig'

    taken from astropy

    .. todo::

        - add docparse tools so can merge docstrings

    """

    def __init__(cls, name: str, bases: T.Any, dct: T.Dict):
        """Set up docstring inheritance."""

        def is_public_member(key):
            return (
                key.startswith("__") and key.endswith("__") and len(key) > 4
            ) or not key.startswith("_")

        # /def

        for key, val in dct.items():
            if (
                (inspect.isfunction(val) or inspect.isdatadescriptor(val))
                and is_public_member(key)
                and val.__doc__ is None
            ):

                for base in cls.__mro__[1:]:
                    super_method = getattr(base, key, None)
                    if super_method is not None:
                        val.__doc__ = super_method.__doc__
                        break
                # /for
        # /for

        super().__init__(name, bases, dct)
        return

    # /def


# /class

##############################################################################
# END



# -*- coding: utf-8 -*-

"""Added functionality to `functools`.

Routine Listings
----------------
`makeFunction`
    make a function from an existing code object.

`copy_function`
    Copy a function.

`update_wrapper`
    this overrides the default ``functools`` `update_wrapper`
    and adds signature and docstring overriding

`wraps`
    overrides the default ``functools`` `update_wrapper`
    and adds signature and docstring overriding

References
----------
Some functions modified from https://docs.python.org/3/library/functools.html

Notes
-----
.. todo::

    improve `makeFunction` call signature
    test `makeFunction`-made function speeds
    todo make partial respect signature

"""


__all__ = [
    "make_function",
    "copy_function",
    "update_wrapper",
    "wraps",
]


##############################################################################
# IMPORTS

# BUILT-IN
import typing as T
from functools import *  # noqa # so can be a drop-in for `functools`
from functools import partial
from types import CodeType, FunctionType

# PROJECT-SPECIFIC
from . import inspect as _nspct
from .inspect import FullerSignature as _FullerSig
from .string import FormatTemplate as _FormatTemplate
from utilipy.extern.doc_parse_tools import store as _store

# CLEAN MULTIPLE DEFINITIONS
del globals()["update_wrapper"]
del globals()["wraps"]


###############################################################################
# CODE
###############################################################################


def make_function(
    code: CodeType,
    globals_: T.Any,
    name: str,
    signature: _FullerSig,
    docstring: str = None,
    closure: T.Any = None,
    qualname: str = None,
    # options
    _add_signature: bool = False,
):
    r"""Make a function with a specified signature and docstring.

    This is pure python and may not make the fastest functions

    Parameters
    ----------
    code : code
        the .__code__ method of a function
    globals_ : Any
    name : str
    signature : Signature
        inspect.Signature converted to utilipy.Signature
    docstring : str
    closure : Any
    qualname : str

    Returns
    -------
    function: Callable
        the created function

    Other Parameters
    ----------------
    \_add_signature : bool
        Whether to add `signature` as ``__signature__``.

    .. todo::

        check how signature and closure relate
        __qualname__

    """
    if not isinstance(signature, _FullerSig):  # not my custom signature
        signature = _FullerSig(
            parameters=signature.parameters.values(),
            return_annotation=signature.return_annotation,
            # docstring=docstring  # not yet implemented
        )
    # else:
    #     pass

    # make function
    function = FunctionType(
        code, globals_, name=name, argdefs=signature.defaults, closure=closure
    )

    # assign properties not (properly) handled by FunctionType
    function.__kwdefaults__ = signature.__kwdefaults__
    function.__annotations__ = signature.__annotations__
    function.__doc__ = docstring

    if qualname is not None:
        function.__qualname__ = qualname

    if _add_signature:
        function.__signature__ = signature.__signature__  # classical signature

    return function


# /def


# -----------------------------------------------------------------------------


def copy_function(func: T.Callable):
    """Copy an existing function.

    Notes
    -----
    copy's code, globals, name, signature, (kw)defaults, docstring, and dict
    custom methods / method overwriting may not be copied

    """
    function = make_function(
        func.__code__,
        func.__globals__,
        name=func.__name__,
        signature=_nspct.fuller_signature(func),
        docstring=func.__doc__,
        closure=func.__closure__,
        qualname=func.__qualname__,
        # options
        _add_signature=hasattr(func, "__signature__"),
    )

    # TODO necessary?
    function.__dict__.update(func.__dict__)
    function.__defaults__ = func.__defaults__
    function.__kwdefaults__ = func.__kwdefaults__

    return function


# /def


###############################################################################
# update_wrapper() and wraps() decorator
###############################################################################

WRAPPER_ASSIGNMENTS = (
    "__module__",
    "__name__",
    "__qualname__",
    "__doc__",
    "__annotations__",
)
SIGNATURE_ASSIGNMENTS = ("__kwdefaults__", "__annotations__")
WRAPPER_UPDATES = ("__dict__",)


def __parse_sig_for_update_wrapper(
    signature: T.Union[_FullerSig, None, bool], wrapped: T.Callable
):
    """Parse signature for `~update_wrapper`.

    Parameters
    ----------
    signature : `~utilipy.utils.inspect.Signature` or None or bool

    Returns
    -------
    signature : `~utilipy.utils.inspect.Signature` or None or False
        If signature is True, returns `~utilipy.utils.inspect.Signature`

    """
    if signature is True:
        _update_sig = True
        signature = _FullerSig.from_callable(wrapped)
    elif signature in (None, False):
        pass
    else:  # convert to my signature object
        _update_sig = False
        # if not (type(signature) == _FullerSig):
        if not isinstance(signature, _FullerSig):
            signature = _FullerSig(
                parameters=signature.parameters.values(),
                return_annotation=signature.return_annotation,
                # docstring=docstring  # not yet implemented
            )
        elif isinstance(signature, _FullerSig):
            pass  # docstring considerations not yet implemented
        else:
            raise ValueError("signature must be a Signature object")
    # /if

    return signature, _update_sig


# /def


def __update_wrapper_update_sig(
    signature: T.Union[_FullerSig, None, bool],
    wrapper_sig: _FullerSig,
    _doc_fmt: T.Optional[dict],
) -> T.Callable:
    """Update signature for `~update_wrapper`."""
    # go through parameters in wrapper_sig, merging into signature
    for param in wrapper_sig.parameters.values():
        # skip _nspct.VAR_POSITIONAL and _nspct.VAR_KEYWORD
        if param.kind in {_nspct.VAR_POSITIONAL, _nspct.VAR_KEYWORD}:
            pass
        # already exists -> replace
        elif param.name in signature.parameters:
            # ensure kind matching
            if param.kind != signature.parameters[param.name].kind:
                raise TypeError(
                    f"{param.name} must match kind in function signature"
                )
            # can only merge keyword-only
            if param.kind == _nspct.KEYWORD_ONLY:
                signature = signature.modify_parameter(
                    param.name,
                    name=None,
                    kind=None,  # inherit b/c matching
                    default=param.default,
                    annotation=param.annotation,
                )

                # track for docstring
                _doc_fmt[param.name] = param.default

        # add to signature
        else:
            # can only merge keyword-only
            if param.kind == _nspct.KEYWORD_ONLY:
                signature = signature.insert_parameter(
                    signature.index_end_keyword_only, param
                )

                # track for docstring
                _doc_fmt[param.name] = param.default
    # /for

    return signature


# /def


def __update_wrapper_docstring(
    wrapped: T.Callable,
    docstring: T.Union[str, bool],
    wrapper_doc: str,
    _doc_style: T.Union[str, T.Callable, None],
):
    """Set docstring for `~update_wrapper`.

    Parameters
    ----------
    docstring : str or bool
    wrapper_doc : strr
    _doc_style : str or Callable or None

    Returns
    -------
    docstring : str

    """
    if isinstance(docstring, str):  # assign docstring
        docstring = _nspct.cleandoc(docstring)
    elif docstring is False:  # just inherit
        docstring = wrapped.__doc__
    else:  # merge wrapper docstring
        wrapped_doc = _nspct.getdoc(wrapped) or ""
        wrapper_doc = _nspct.cleandoc(wrapper_doc)

        if _doc_style is None:  # use original wrapper docstring
            docstring = wrapped_doc + "\n\n" + wrapper_doc

        else:  # TODO implement the full set of options
            docstring = _store[_doc_style](
                wrapped_doc,
                wrapper_doc,
                method="merge",
            )
    # /if

    return docstring


# /def


def update_wrapper(
    wrapper: T.Callable,
    wrapped: T.Callable,
    signature: T.Union[_FullerSig, None, bool] = True,  # not in functools
    docstring: T.Union[str, bool] = True,  # not in functools
    assigned: T.Sequence[str] = WRAPPER_ASSIGNMENTS,
    updated: T.Sequence[str] = WRAPPER_UPDATES,
    # docstring options
    _doc_fmt: T.Optional[dict] = None,  # not in functools
    _doc_style: T.Union[str, T.Callable, None] = None,
):
    """Update a wrapper function to look like the wrapped function.

    Parameters
    ----------
    wrapper : Callable
        the function to be updated
    wrapped : Callable
       the original function
    signature : Signature or None or bool, optional
        signature to impose on `wrapper`.
        None and False default to `wrapped`'s signature.
        True merges `wrapper` and `wrapped` kwdefaults & annotations
    docstring : str or bool, optional
        docstring to impose on `wrapper`.
        False ignores `wrapper`'s docstring, using only `wrapped`'s docstring.
        None (defualt) merges the `wrapper` and `wrapped` docstring
    assigned : tuple, optional
       tuple naming the attributes assigned directly
       from the wrapped function to the wrapper function (defaults to
       ``functools.WRAPPER_ASSIGNMENTS``)
    updated : tuple, optional
       is a tuple naming the attributes of the wrapper that
       are updated with the corresponding attribute from the wrapped
       function (defaults to ``functools.WRAPPER_UPDATES``)
    _doc_fmt : dict, optional
        dictionary to format wrapper docstring
    _doc_style: str or Callable, optional
        the style of the docstring
        if None (default), appends `wrapper` docstring
        if str or Callable, merges the docstring

    Returns
    -------
    wrapper : Callable
        `wrapper` function updated by the `wrapped` function's attributes and
        also the provided `signature` and `docstring`.

    Raises
    ------
    ValueError
        if docstring is True

    """
    # ---------------------------------------
    # preamble

    signature, _update_sig = __parse_sig_for_update_wrapper(signature, wrapped)

    # need to get wrapper properties now
    wrapper_sig = _FullerSig.from_callable(wrapper)

    wrapper_doc = _nspct.getdoc(wrapper) or ""
    wrapper_doc = "\n".join(wrapper_doc.split("\n")[1:])  # drop title

    if _doc_fmt is None:
        _doc_fmt = {}

    # ---------------------------------------
    # update wrapper (same as functools.update_wrapper)

    for attr in assigned:
        try:
            value = getattr(wrapped, attr)
        except AttributeError:
            pass
        else:
            setattr(wrapper, attr, value)

    for attr in updated:  # update whole dictionary
        getattr(wrapper, attr).update(getattr(wrapped, attr, {}))

    # ---------------------------------------

    # deal with signature
    if signature in (None, False):
        pass

    elif _update_sig:  # merge wrapped and wrapper signature

        signature = __update_wrapper_update_sig(
            signature, wrapper_sig, _doc_fmt
        )

        for attr in SIGNATURE_ASSIGNMENTS:
            value = getattr(signature, attr)
            setattr(wrapper, attr, value)

        wrapper.__signature__ = signature.signature

    else:  # a signature object
        for attr in SIGNATURE_ASSIGNMENTS:
            _value = getattr(signature, attr)
            setattr(wrapper, attr, _value)

        # for docstring
        for param in wrapper_sig.parameters.values():
            # can only merge keyword-only
            if param.kind == _nspct.KEYWORD_ONLY:
                _doc_fmt[param.name] = param.default

        wrapper.__signature__ = signature.signature

    # ---------------------------------------
    # docstring

    if _doc_fmt:  # (not empty dict)
        wrapper_doc = _FormatTemplate(wrapper_doc).safe_substitute(**_doc_fmt)

    wrapper.__doc__ = __update_wrapper_docstring(
        wrapped,
        docstring=docstring,
        wrapper_doc=wrapper_doc,
        _doc_style=_doc_style,
    )

    # Issue #17482: set __wrapped__ last so we don't inadvertently copy it
    # from the wrapped function when updating __dict__
    wrapper.__wrapped__ = wrapped
    # Return the wrapper so this can be used as a decorator via partial()
    return wrapper


# /def


# -----------------------------------------------------------------------------


def wraps(
    wrapped: T.Callable,
    signature: T.Union[_FullerSig, None, bool] = True,
    docstring: T.Union[str, None, bool] = None,
    assigned: T.Sequence[str] = WRAPPER_ASSIGNMENTS,
    updated: T.Sequence[str] = WRAPPER_UPDATES,
    _doc_fmt: T.Optional[dict] = None,
    _doc_style: T.Union[str, T.Callable, None] = None,
):
    """:func:`~functools.wraps`, adding signature and docstring features.

    Decorator factory to apply ``update_wrapper()`` to a wrapper function.

    This is a convenience function to simplify applying ``partial()`` to
    ``update_wrapper()``.

    Parameters
    ----------
    wrapped: Callable
    signature: _FullerSig or bool or None, optional
        True (default)
    docstring: str or bool or None, optional
        None
    assigned: Sequence[str]
        WRAPPER_ASSIGNMENTS,
    updated: Sequence[str]
        WRAPPER_UPDATES,
    _doc_fmt: dict or None, optional
    _doc_style: Union[str, Callable, None], optional

    Returns
    -------
    partial
        a decorator that invokes ``update_wrapper()`` with the decorated
        function as the wrapper argument and the arguments to ``wraps()`` as
        the remaining arguments. Default arguments are as for
        ``update_wrapper()``.

    """
    return partial(
        update_wrapper,
        wrapped=wrapped,
        signature=signature,
        docstring=docstring,
        assigned=assigned,
        updated=updated,
        _doc_fmt=_doc_fmt,
        _doc_style=_doc_style,
    )


# /def


##############################################################################
# END



# -*- coding: utf-8 -*-

"""Functions for use in testing :mod:`~utilipy.utils.functools`."""

__author__ = "Nathaniel Starkman"

__all__ = [
    # args
    "func_x",
    "func_xy",
    "func_xy_annotate",
    # defargs
    "func_a",
    "func_ab",
    "func_ab_annotate",
    # *args
    "func_args",
    "func_args_annotate",
    # kwargs
    "func_p",
    "func_pq",
    "func_pq_annotate",
    # **kwargs
    "func_kwargs",
    "func_kwargs_annotate",
    # args, defargs
    "func_x_a",
    "func_xy_ab",
    "func_xy_ab_annotate",
    # args, *args
    "func_x_args",
    "func_xy_args",
    "func_xy_args_annotate",
    # args, kwargs
    "func_x_p",
    "func_xy_pq",
    "func_xy_pq_annotate",
    # args, **kwargs
    "func_x_kwargs",
    "func_xy_kwargs",
    "func_xy_kwargs_annotate",
    # defargs, *args
    "func_a_args",
    "func_ab_args",
    "func_ab_args_annotate",
    # defargs, kwargs
    "func_a_p",
    "func_ab_pq",
    "func_ab_pq_annotate",
    # defargs, **kwargs
    "func_a_kwargs",
    "func_ab_kwargs",
    "func_ab_kwargs_annotate",
    # *args, kwargs
    "func_args_p",
    "func_args_pq",
    "func_args_pq_annotate",
    # *args, **kwargs
    "func_args_kwargs",
    "func_args_kwargs_annotate",
    # args, defargs, **args
    "func_x_a_args",
    "func_xy_ab_args",
    "func_xy_ab_args_annotate",
    # args, defargs, kwargs
    "func_x_a_p",
    "func_xy_ab_pq",
    "func_xy_ab_pq_annotate",
    # args, defargs, **kwargs
    "func_x_a_kwargs",
    "func_xy_ab_kwargs",
    "func_xy_ab_kwargs_annotate",
    # args, *args, kwargs
    "func_x_args_p",
    "func_xy_args_pq",
    "func_xy_args_pq_annotate",
    # args, *args, **kwargs
    "func_x_args_kwargs",
    "func_xy_args_kwargs",
    "func_xy_args_kwargs_annotate",
    # defargs, *args, kwargs
    "func_a_args_p",
    "func_ab_args_pq",
    "func_ab_args_pq_annotate",
    # defargs, *args, **kwargs
    "func_a_args_kwargs",
    "func_ab_args_kwargs",
    "func_ab_args_kwargs_annotate",
    # *args, kwargs, **kwargs
    "func_args_p_kwargs",
    "func_args_pq_kwargs",
    "func_args_pq_kwargs_annotate",
    # args, defargs, *args, kwargs
    "func_x_a_args_p",
    "func_xy_ab_args_pq",
    "func_xy_ab_args_pq_annotate",
    # args, defargs, *args, **kwargs
    "func_x_a_args_kwargs",
    "func_xy_ab_args_kwargs",
    "func_xy_ab_args_kwargs_annotate",
    # args, defargs, kwargs, **kwargs
    "func_x_a_p_kwargs",
    "func_xy_ab_pq_kwargs",
    "func_xy_ab_pq_kwargs_annotate",
    # args, *args, kwargs, **kwargs
    "func_x_args_p_kwargs",
    "func_xy_args_pq_kwargs",
    "func_xy_args_pq_kwargs_annotate",
    # defargs, *args, kwargs, **kwargs
    "func_a_args_p_kwargs",
    "func_ab_args_pq_kwargs",
    "func_ab_args_pq_kwargs_annotate",
    # args, defargs, *args, kwargs, **kwargs
    "func_x_a_args_p_kwargs",
    "func_xy_ab_args_pq_kwargs",
    "func_xy_ab_args_pq_kwargs_annotate",
]

##############################################################################
# IMPORTS

# BUILT-IN
import typing

##############################################################################
# CODE
##############################################################################

# -----------------------------------------------------------------------------
# Make Function NameSpace

# ----------------------------------------------
# Function with args


def func_x(x):
    """func.

    Parameters
    ----------
    x: float

    Returns
    -------
    x: float
    None, None, None, None, None, None, None

    """
    return x, None, None, None, None, None, None, None


# /def


def func_xy(x, y):
    """func.

    Parameters
    ----------
    x, y: float

    Returns
    -------
    x, y: float
    None, None, None, None, None, None

    """
    return x, y, None, None, None, None, None, None


# /def


@typing.no_type_check
def func_xy_annotate(x: "0", y) -> typing.Tuple:
    """func.

    Parameters
    ----------
    x, y: float

    Returns
    -------
    x, y: float
    None, None, None, None, None, None

    """
    return x, y, None, None, None, None, None, None


# /def


# ----------------------------------------------
# Function with defargs


def func_a(a=2):
    """func.

    Parameters
    ----------
    a: int, optional

    Returns
    -------
    None, None
    a: int
    None, None, None, None, None

    """
    return None, None, a, None, None, None, None, None


# /def


def func_ab(a=2, b=3):
    """func.

    Parameters
    ----------
    a, b: int, optional

    Returns
    -------
    None, None
    a, b: int
    None, None, None, None

    """
    return None, None, a, b, None, None, None, None


# /def


@typing.no_type_check
def func_ab_annotate(a: "1" = 2, b=3) -> typing.Tuple:
    """func.

    Parameters
    ----------
    a, b: int, optional

    Returns
    -------
    None, None
    a, b: int
    None, None, None, None

    """
    return None, None, a, b, None, None, None, None


# /def


# ----------------------------------------------
# Function with *args


def func_args(*args):
    """func.

    Parameters
    ----------
    *args

    Returns
    -------
    None, None, None, None
    args: tuple
    None, None, None

    """
    return None, None, None, None, args, None, None, None


# /def


@typing.no_type_check
def func_args_annotate(*args: "2") -> typing.Tuple:
    """func.

    Parameters
    ----------
    *args

    Returns
    -------
    None, None, None, None
    args: tuple
    None, None, None

    """
    return None, None, None, None, args, None, None, None


# /def


# ----------------------------------------------
# Function with kwargs


def func_p(*, p="p"):
    """func.

    Parameters
    ----------
    p: str, optional

    Returns
    -------
    p: str

    """
    return None, None, None, None, None, p, None, None


# /def


def func_pq(*, p="p", q="q"):
    """func.

    Parameters
    ----------
    p, q: str, optional

    Returns
    -------
    p, q: str

    """
    return None, None, None, None, None, p, q, None


# /def


@typing.no_type_check
def func_pq_annotate(*, p: "3" = "p", q="q") -> typing.Tuple:
    """func.

    Parameters
    ----------
    p, q: str, optional

    Returns
    -------
    p, q: str

    """
    return None, None, None, None, None, p, q, None


# /def


# ----------------------------------------------
# Function with **kwargs


def func_kwargs(**kwargs):
    """func.

    Parameters
    ----------
    **kwargs

    Returns
    -------
    kwargs: dict

    """
    return None, None, None, None, None, None, None, kwargs


# /def


@typing.no_type_check
def func_kwargs_annotate(**kwargs: "4") -> typing.Tuple:
    """func.

    Parameters
    ----------
    **kwargs

    Returns
    -------
    kwargs: dict

    """
    return None, None, None, None, None, None, None, kwargs


# /def


# -----------------------------------------------------------------------------
# Function with args, kwargs


def func_x_a(x, a=2):
    """func.

    Parameters
    ----------
    x: float
    a: int, optional

    Returns
    -------
    x: float
    a: int

    """
    return x, None, a, None, None, None, None, None


# /def


def func_xy_ab(x, y, a=2, b=3):
    """func.

    Parameters
    ----------
    x, y: float
    a, b: int, optional

    Returns
    -------
    x, y: float
    a, b: int

    """
    return x, y, a, b, None, None, None, None


# /def


@typing.no_type_check
def func_xy_ab_annotate(x: "0", y, a: "1" = 2, b=3) -> typing.Tuple:
    """func.

    Parameters
    ----------
    x, y: float
    a, b: int, optional

    Returns
    -------
    x, y: float
    a, b: int

    """
    return x, y, a, b, None, None, None, None


# /def


# ----------------------------------------------
# Function with args, *args


def func_x_args(x, *args):
    """func.

    Parameters
    ----------
    x: float
    args: tuple

    Returns
    -------
    x: float
    args: tuple

    """
    return x, None, None, None, args, None, None, None


# /def


def func_xy_args(x, y, *args):
    """func.

    Parameters
    ----------
    x, y: float
    args: tuple

    Returns
    -------
    x, y: float
    args: tuple

    """
    return x, y, None, None, args, None, None, None


# /def


@typing.no_type_check
def func_xy_args_annotate(x: "0", y, *args: "2") -> typing.Tuple:
    """func.

    Parameters
    ----------
    x, y: float
    args: tuple

    Returns
    -------
    x, y: float
    args: tuple

    """
    return x, y, None, None, args, None, None, None


# /def


# ----------------------------------------------
# Function with args, kwargs


def func_x_p(x, *, p="p"):
    """func.

    Parameters
    ----------
    x: float
    p: str, optional

    Returns
    -------
    x: float
    p: str

    """
    return x, None, None, None, None, p, None, None


# /def


def func_xy_pq(x, y, *, p="p", q="q"):
    """func.

    Parameters
    ----------
    x, y: float
    p, q: str, optional

    Returns
    -------
    x, y: float
    p, q: str

    """
    return x, y, None, None, None, p, q, None


# /def


@typing.no_type_check
def func_xy_pq_annotate(x: "0", y, *, p: "3" = "p", q="q") -> typing.Tuple:
    """func.

    Parameters
    ----------
    x, y: float
    p, q: str, optional

    Returns
    -------
    x, y: float
    p, q: str

    """
    return x, y, None, None, None, p, q, None


# /def


# ----------------------------------------------
# Function with args, **kwargs


def func_x_kwargs(x, **kwargs):
    """func.

    Parameters
    ----------
    x: float
    kwargs: dict

    Returns
    -------
    x: float
    kwargs: dict

    """
    return x, None, None, None, None, None, None, kwargs


# /def


def func_xy_kwargs(x, y, **kwargs):
    """func.

    Parameters
    ----------
    x, y: float
    kwargs: dict

    Returns
    -------
    x, y: float
    kwargs: dict

    """
    return x, y, None, None, None, None, None, kwargs


# /def


@typing.no_type_check
def func_xy_kwargs_annotate(x: "0", y, **kwargs: "4") -> typing.Tuple:
    """func.

    Parameters
    ----------
    x, y: float
    kwargs: dict

    Returns
    -------
    x, y: float
    kwargs: dict

    """
    return x, y, None, None, None, None, None, kwargs


# /def


# ----------------------------------------------
# Function with defargs, *args


def func_a_args(a=2, *args):
    """func.

    Parameters
    ----------
    a: int
    args: tuple

    Returns
    -------
    a: int
    args: tuple

    """
    return None, None, a, None, args, None, None, None


# /def


def func_ab_args(a=2, b=3, *args):
    """func.

    Parameters
    ----------
    a, b: int
    args: tuple

    Returns
    -------
    a, b: int
    args: tuple

    """
    return None, None, a, b, args, None, None, None


# /def


@typing.no_type_check
def func_ab_args_annotate(a: "1" = 2, b=3, *args: "2") -> typing.Tuple:
    """func.

    Parameters
    ----------
    a, b: int
    args: tuple

    Returns
    -------
    a, b: int
    args: tuple

    """
    return None, None, a, b, args, None, None, None


# /def


# ----------------------------------------------
# Function with defargs, kwargs


def func_a_p(a=2, *, p="p"):
    """func.

    Parameters
    ----------
    a: int
    p: str, optional

    Returns
    -------
    a: int
    p: str

    """
    return None, None, a, None, p, None, None, None


# /def


def func_ab_pq(a=2, b=3, *, p="p", q="q"):
    """func.

    Parameters
    ----------
    a, b: int
    p, q: str, optional

    Returns
    -------
    a, b: int
    p, q: str

    """
    return None, None, a, b, None, p, q, None


# /def


@typing.no_type_check
def func_ab_pq_annotate(
    a: "1" = 2, b=3, *, p: "3" = "p", q="q"
) -> typing.Tuple:
    """func.

    Parameters
    ----------
    a, b: int
    p, q: str, optional

    Returns
    -------
    a, b: int
    p, q: str

    """
    return None, None, a, b, None, p, q, None


# /def


# ----------------------------------------------
# Function with defargs, **kwargs


def func_a_kwargs(a=2, **kwargs):
    """func.

    Parameters
    ----------
    a: int, optional
    kwargs: dict

    Returns
    -------
    a: int
    kwargs: dict

    """
    return None, None, a, None, None, None, None, kwargs


# /def


def func_ab_kwargs(a=2, b=3, **kwargs):
    """func.

    Parameters
    ----------
    a, b: int, optional
    kwargs: dict

    Returns
    -------
    a, b: int
    kwargs: dict

    """
    return None, None, a, b, None, None, None, kwargs


# /def


@typing.no_type_check
def func_ab_kwargs_annotate(a: "1" = 2, b=3, **kwargs: "4") -> typing.Tuple:
    """func.

    Parameters
    ----------
    a, b: int, optional
    kwargs: dict

    Returns
    -------
    a, b: int
    kwargs: dict

    """
    return None, None, a, b, None, None, None, kwargs


# /def


# ----------------------------------------------
# Function with *args, kwargs


def func_args_p(*args, p="p"):
    """func.

    Parameters
    ----------
    args: tuple
    p: str, optional

    Returns
    -------
    args: tuple
    p: str

    """
    return None, None, None, None, args, p, None, None


# /def


def func_args_pq(*args, p="p", q="q"):
    """func.

    Parameters
    ----------
    args: tuple
    p, q: str, optional

    Returns
    -------
    args: tuple
    p, q: str

    """
    return None, None, None, None, args, p, q, None


# /def


@typing.no_type_check
def func_args_pq_annotate(*args: "2", p: "3" = "p", q="q") -> typing.Tuple:
    """func.

    Parameters
    ----------
    args: tuple
    p, q: str, optional

    Returns
    -------
    args: tuple
    p, q: str

    """
    return None, None, None, None, args, p, q, None


# /def


# ----------------------------------------------
# Function with *args, **kwargs


def func_args_kwargs(*args, **kwargs):
    """func.

    Parameters
    ----------
    args: tuple
    kwargs: dict

    Returns
    -------
    args: tuple
    kwargs: dict

    """
    return None, None, None, None, args, None, None, kwargs


# /def


@typing.no_type_check
def func_args_kwargs_annotate(*args: "2", **kwargs: "4") -> typing.Tuple:
    """func.

    Parameters
    ----------
    args: tuple
    kwargs: dict

    Returns
    -------
    args: tuple
    kwargs: dict

    """
    return None, None, None, None, args, None, None, kwargs


# /def


# -----------------------------------------------------------------------------
# Function with args, defargs, *args


def func_x_a_args(x, a=2, *args):
    """func.

    Parameters
    ----------
    x: float
    a: int
    args: tuple

    Returns
    -------
    x: float
    a: int
    args: tuple

    """
    return x, None, a, None, args, None, None, None


# /def


def func_xy_ab_args(x, y, a=2, b=3, *args):
    """func.

    Parameters
    ----------
    x, y: float
    a, b: int
    args: tuple

    Returns
    -------
    x, y: float
    a, b: int
    args: tuple

    """
    return x, y, a, b, args, None, None, None


# /def


@typing.no_type_check
def func_xy_ab_args_annotate(
    x: "0", y, a: "1" = 2, b=3, *args: "2"
) -> typing.Tuple:
    """func.

    Parameters
    ----------
    x, y: float
    a, b: int
    args: tuple

    Returns
    -------
    x, y: float
    a, b: int
    args: tuple

    """
    return x, y, a, b, args, None, None, None


# /def


# ----------------------------------------------
# Function with args, defargs, kwargs


def func_x_a_p(x, a=2, *, p="p"):
    """func.

    Parameters
    ----------
    x: float
    a: int
    p: str

    Returns
    -------
    x: float
    a: int
    p: str

    """
    return x, None, a, None, None, p, None, None


# /def


def func_xy_ab_pq(x, y, a=2, b=3, *, p="p", q="q"):
    """func.

    Parameters
    ----------
    x, y: float
    a, b: int
    p, q: str

    Returns
    -------
    x, y: float
    a, b: int
    p, q: str

    """
    return x, y, a, b, None, p, q, None


# /def


@typing.no_type_check
def func_xy_ab_pq_annotate(
    x: "0", y, a: int = 2, b=3, *, p: "3" = "p", q="q"
) -> typing.Tuple:
    """func.

    Parameters
    ----------
    x, y: float
    a, b: int
    p, q: str

    Returns
    -------
    x, y: float
    a, b: int
    p, q: str

    """
    return x, y, a, b, None, p, q, None


# /def


# ----------------------------------------------
# Function with args, defargs, **kwargs


def func_x_a_kwargs(x, a=2, **kwargs):
    """func.

    Parameters
    ----------
    x: float
    a: int
    kwargs: dict

    Returns
    -------
    x: float
    a: int
    kwargs: dict

    """
    return x, None, a, None, None, None, None, kwargs


# /def


def func_xy_ab_kwargs(x, y, a=2, b=3, **kwargs):
    """func.

    Parameters
    ----------
    x, y: float
    a, b: int
    kwargs: dict

    Returns
    -------
    x, y: float
    a, b: int
    kwargs: dict

    """
    return x, y, a, b, None, None, None, kwargs


# /def


@typing.no_type_check
def func_xy_ab_kwargs_annotate(
    x: "0", y, a: int = 2, b=3, **kwargs: "4"
) -> typing.Tuple:
    """func.

    Parameters
    ----------
    x, y: float
    a, b: int
    kwargs: dict

    Returns
    -------
    x, y: float
    a, b: int
    kwargs: dict

    """
    return x, y, a, b, None, None, None, kwargs


# /def


# ----------------------------------------------
# Function with args, *args, kwargs


def func_x_args_p(x, *args, p="p"):
    """func.

    Parameters
    ----------
    x: float
    args: tuple
    p: str

    Returns
    -------
    x: float
    args: tuple
    p: str

    """
    return x, None, None, None, args, p, None, None


# /def


def func_xy_args_pq(x, y, *args, p="p", q="q"):
    """func.

    Parameters
    ----------
    x, y: float
    args: tuple
    p, q: str

    Returns
    -------
    x, y: float
    args: tuple
    p, q: str

    """
    return x, y, None, None, args, p, q, None


# /def


@typing.no_type_check
def func_xy_args_pq_annotate(
    x: "0", y, *args: "2", p: "3" = "p", q="q"
) -> typing.Tuple:
    """func.

    Parameters
    ----------
    x, y: float
    args: tuple
    p, q: str

    Returns
    -------
    x, y: float
    args: tuple
    p, q: str

    """
    return x, y, None, None, args, p, q, None


# /def


# ----------------------------------------------
# Function with args, *args, **kwargs


def func_x_args_kwargs(x, *args, **kwargs):
    """func.

    Parameters
    ----------
    x: float
    args: tuple
    kwargs: dict

    Returns
    -------
    x: float
    args: tuple
    kwargs: dict

    """
    return x, None, None, None, args, None, None, kwargs


# /def


def func_xy_args_kwargs(x, y, *args, **kwargs):
    """func.

    Parameters
    ----------
    x, y: float
    args: tuple
    kwargs: dict

    Returns
    -------
    x, y: float
    args: tuple
    kwargs: dict

    """
    return x, y, None, None, args, None, None, kwargs


# /def


@typing.no_type_check
def func_xy_args_kwargs_annotate(
    x: "0", y, *args: "2", **kwargs: "4"
) -> typing.Tuple:
    """func.

    Parameters
    ----------
    x, y: float
    args: tuple
    kwargs: dict

    Returns
    -------
    x, y: float
    args: tuple
    kwargs: dict

    """
    return x, y, None, None, args, None, None, kwargs


# /def


# ----------------------------------------------
# Function with defargs, *args, kwargs


def func_a_args_p(a=2, *args, p="p"):
    """func.

    Parameters
    ----------
    a: int
    args: tuple
    p: str

    Returns
    -------
    a: int
    args: tuple
    p: str

    """
    return None, None, a, None, args, p, None, None


# /def


def func_ab_args_pq(a=2, b=3, *args, p="p", q="q"):
    """func.

    Parameters
    ----------
    a, b: int
    args: tuple
    p, q: str

    Returns
    -------
    a, b: int
    args: tuple
    p, q: str

    """
    return None, None, a, b, args, p, q, None


# /def


@typing.no_type_check
def func_ab_args_pq_annotate(
    a: "1" = 2, b=3, *args: "2", p: "3" = "p", q="q"
) -> typing.Tuple:
    """func.

    Parameters
    ----------
    a, b: int
    args: tuple
    p, q: str

    Returns
    -------
    a, b: int
    args: tuple
    p, q: str

    """
    return None, None, a, b, args, p, q, None


# /def


# ----------------------------------------------
# Function with defargs, *args, **kwargs


def func_a_args_kwargs(a=2, *args, **kwargs):
    """func.

    Parameters
    ----------
    a: int
    args: tuple
    kwargs: dict

    Returns
    -------
    a: int
    args: tuple
    kwargs: dict

    """
    return None, None, a, None, args, None, None, kwargs


# /def


def func_ab_args_kwargs(a=2, b=3, *args, **kwargs):
    """func.

    Parameters
    ----------
    a, b: int
    args: tuple
    kwargs: dict

    Returns
    -------
    a, b: int
    args: tuple
    kwargs: dict

    """
    return None, None, a, b, args, None, None, kwargs


# /def


@typing.no_type_check
def func_ab_args_kwargs_annotate(
    a: "1" = 2, b=3, *args: "2", **kwargs: "4"
) -> typing.Tuple:
    """func.

    Parameters
    ----------
    a, b: int
    args: tuple
    kwargs: dict

    Returns
    -------
    a, b: int
    args: tuple
    kwargs: dict

    """
    return None, None, a, b, args, None, None, kwargs


# /def


# ----------------------------------------------
# Function with *args, kwargs, **kwargs


def func_args_p_kwargs(*args, p="p", **kwargs):
    """func.

    Parameters
    ----------
    args: tuple
    p: str
    kwargs: dict

    Returns
    -------
    args: tuple
    p: str
    kwargs: dict

    """
    return None, None, None, None, args, p, None, kwargs


# /def


def func_args_pq_kwargs(*args, p="p", q="q", **kwargs):
    """func.

    Parameters
    ----------
    args: tuple
    p, q: str
    kwargs: dict

    Returns
    -------
    args: tuple
    p, q: str
    kwargs: dict

    """
    return None, None, None, None, args, p, q, kwargs


# /def


@typing.no_type_check
def func_args_pq_kwargs_annotate(
    *args: "2", p: "3" = "p", q="q", **kwargs: "4"
) -> typing.Tuple:
    """func.

    Parameters
    ----------
    args: tuple
    p, q: str
    kwargs: dict

    Returns
    -------
    args: tuple
    p, q: str
    kwargs: dict

    """
    return None, None, None, None, args, p, q, kwargs


# /def


# ----------------------------------------------
# Function with args, defargs, *args, kwargs


def func_x_a_args_p(x, a=2, *args, p="p"):
    """func.

    Parameters
    ----------
    x: float
    a: int
    args: tuple
    p: str

    Returns
    -------
    x: float
    a: int
    args: tuple
    p: str

    """
    return x, None, a, None, args, p, None, None


# /def


def func_xy_ab_args_pq(x, y, a=2, b=3, *args, p="p", q="q"):
    """func.

    Parameters
    ----------
    x, y: float
    a, b: int
    args: tuple
    p, q: str

    Returns
    -------
    x, y: float
    a, b: int
    args: tuple
    p, q: str

    """
    return x, y, a, b, args, p, q, None


# /def


@typing.no_type_check
def func_xy_ab_args_pq_annotate(
    x: "0", y, a: "1" = 2, b=3, *args: "2", p: "3" = "p", q="q"
) -> typing.Tuple:
    """func.

    Parameters
    ----------
    x, y: float
    a, b: int
    args: tuple
    p, q: str

    Returns
    -------
    x, y: float
    a, b: int
    args: tuple
    p, q: str

    """
    return x, y, a, b, args, p, q, None


# /def


# ----------------------------------------------
# Function with args, defargs, *args, **kwargs


def func_x_a_args_kwargs(x, a=2, *args, **kwargs):
    """func.

    Parameters
    ----------
    x: float
    a: int
    args: tuple
    kwargs: dict

    Returns
    -------
    x: float
    a: int
    args: tuple
    kwargs: dict

    """
    return x, None, a, None, args, None, None, kwargs


# /def


def func_xy_ab_args_kwargs(x, y, a=2, b=3, *args, **kwargs):
    """func.

    Parameters
    ----------
    x, y: float
    a, b: int
    args: tuple
    kwargs: dict

    Returns
    -------
    x, y: float
    a, b: int
    args: tuple
    kwargs: dict

    """
    return x, y, a, b, args, None, None, kwargs


# /def


@typing.no_type_check
def func_xy_ab_args_kwargs_annotate(
    x: "0", y, a: "1" = 2, b=3, *args: "2", **kwargs: "4"
) -> typing.Tuple:
    """func.

    Parameters
    ----------
    x, y: float
    a, b: int
    args: tuple
    kwargs: dict

    Returns
    -------
    x, y: float
    a, b: int
    args: tuple
    kwargs: dict

    """
    return x, y, a, b, args, None, None, kwargs


# /def


# ----------------------------------------------
# Function with args, defargs, kwargs, **kwargs


def func_x_a_p_kwargs(x, a=2, *, p="p", **kwargs):
    """func.

    Parameters
    ----------
    x: float
    a: int
    p: str
    kwargs: dict

    Returns
    -------
    x: float
    a: int
    p: str
    kwargs: dict

    """
    return x, None, a, None, None, p, None, kwargs


# /def


def func_xy_ab_pq_kwargs(x, y, a=2, b=3, *, p="p", q="q", **kwargs):
    """func.

    Parameters
    ----------
    x, y: float
    a, b: int
    p, q: str
    kwargs: dict

    Returns
    -------
    x, y: float
    a, b: int
    p, q: str
    kwargs: dict

    """
    return x, y, a, b, None, p, q, kwargs


# /def


@typing.no_type_check
def func_xy_ab_pq_kwargs_annotate(
    x: "0", y, a: "1" = 2, b=3, *, p: "3" = "p", q="q", **kwargs: "4"
) -> typing.Tuple:
    """func.

    Parameters
    ----------
    x, y: float
    a, b: int
    p, q: str
    kwargs: dict

    Returns
    -------
    x, y: float
    a, b: int
    p, q: str
    kwargs: dict

    """
    return x, y, a, b, None, p, q, kwargs


# /def


# ----------------------------------------------
# Function with args, *args, kwargs, **kwargs


def func_x_args_p_kwargs(x, *args, p="p", **kwargs):
    """func.

    Parameters
    ----------
    x: float
    args: tuple
    p: str
    kwargs: dict

    Returns
    -------
    x: float
    args: tuple
    p: str
    kwargs: dict

    """
    return x, None, None, None, args, p, None, kwargs


# /def


def func_xy_args_pq_kwargs(x, y, *args, p="p", q="q", **kwargs):
    """func.

    Parameters
    ----------
    x, y: float
    args: tuple
    p, q: str
    kwargs: dict

    Returns
    -------
    x, y: float
    args: tuple
    p, q: str
    kwargs: dict

    """
    return x, y, None, None, args, p, q, kwargs


# /def


@typing.no_type_check
def func_xy_args_pq_kwargs_annotate(
    x: "0", y, *args, p: "3" = "p", q="q", **kwargs: "4"
) -> typing.Tuple:
    """func.

    Parameters
    ----------
    x, y: float
    args: tuple
    p, q: str
    kwargs: dict

    Returns
    -------
    x, y: float
    args: tuple
    p, q: str
    kwargs: dict

    """
    return x, y, None, None, args, p, q, kwargs


# /def


# ----------------------------------------------
# Function with defargs, *args, kwargs, **kwargs


def func_a_args_p_kwargs(a=2, *args, p="p", **kwargs):
    """func.

    Parameters
    ----------
    a: int
    args: tuple
    p: str
    kwargs: dict

    Returns
    -------
    a: int
    args: tuple
    p: str
    kwargs: dict

    """
    return None, None, a, None, args, p, None, kwargs


# /def


def func_ab_args_pq_kwargs(a=2, b=3, *args, p="p", q="q", **kwargs):
    """func.

    Parameters
    ----------
    a, b: int
    args: tuple
    p, q: str
    kwargs: dict

    Returns
    -------
    a, b: int
    args: tuple
    p, q: str
    kwargs: dict

    """
    return None, None, a, b, args, p, q, kwargs


# /def


@typing.no_type_check
def func_ab_args_pq_kwargs_annotate(
    a: "1" = 2, b=3, *args: "2", p: "3" = "p", q="q", **kwargs: "4"
) -> typing.Tuple:
    """func.

    Parameters
    ----------
    a, b: int
    args: tuple
    p, q: str
    kwargs: dict

    Returns
    -------
    a, b: int
    args: tuple
    p, q: str
    kwargs: dict

    """
    return None, None, a, b, args, p, q, kwargs


# /def


# ----------------------------------------------
# Function with args, defargs, *args, kwargs, **kwargs


def func_x_a_args_p_kwargs(x, a=2, *args, p="p", **kwargs):
    """func.

    Parameters
    ----------
    x: float
    a: int
    args: tuple
    p: str
    kwargs: dict

    Returns
    -------
    x: float
    a: int
    args: tuple
    p: str
    kwargs: dict

    """
    return x, None, a, None, args, p, None, kwargs


# /def


def func_xy_ab_args_pq_kwargs(x, y, a=2, b=3, *args, p="p", q="q", **kwargs):
    """func.

    Parameters
    ----------
    x, y: float
    a, b: int
    args: tuple
    p, q: str, optional
    kwargs: dict

    Returns
    -------
    x, y: float
    a, b: int
    args: tuple
    p, q: str
    kwargs: dict

    """
    return x, y, a, b, args, p, q, kwargs


# /def


@typing.no_type_check
def func_xy_ab_args_pq_kwargs_annotate(
    x, y, a: "1" = 2, b=3, *args: "2", p: "3" = "p", q="q", **kwargs: "4"
) -> typing.Tuple:
    """func.

    Parameters
    ----------
    x, y: float
    a, b: int
    args: tuple
    p, q: str, optional
    kwargs: dict

    Returns
    -------
    x, y: float
    a, b: int
    args: tuple
    p, q: str
    kwargs: dict

    """
    return x, y, a, b, args, p, q, kwargs


# /def


##############################################################################
# END



# -*- coding: utf-8 -*-
# see LICENSE.rst

"""Data Transformation.

The plan is to build an analog to astropy's TransformGraph, which is only for
coordinate frames, but for converting between arbitrary data types.
The advantage of this kind of system is that many DataTransformGraphs can be
created, each of which is intended for different use. For instance, the xmatch
functions need to find the reference epoch in a data set and ensure that the
information is converted in a particular way. This conversion is need-specific
and would not be present in a very general converter, like PanDoc.

Notes
-----
These are dev notes for my thoughts on building the TransformGraph.

- What about subclasses? how can a subclass be recognized to use a particular
  conversion?
- What about Any-to-something conversions, like trying Table(x) or array(y)?
  these should have lower priority than a registered conversion, but should
  work.

    - have a Any-to-this registry function to add these catch-all converters.
    - Maybe have an option for pre-populating these in the graph?

"""

__author__ = "Nathaniel Starkman"
__credits__ = ["Astropy"]


__all__ = [
    # modules
    "graph",
    "transformations",
    # functions
    "TransformGraph",
    # transformations
    "DataTransform",
    "CompositeTransform",
    # Parameters
    "data_graph",
]


##############################################################################
# IMPORTS

# BUILT IN

# PROJECT-SPECIFIC
from . import graph, transformations
from .graph import TransformGraph
from .transformations import CompositeTransform, DataTransform

##############################################################################
# PARAMETERS

data_graph = TransformGraph(seed_basic=True)


##############################################################################
# END
